# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer

from buildbot.process.results import SUCCESS
from buildbot.test import fakedb


class ReporterTestMixin:
    def setup_reporter_test(self):
        self.reporter_test_project = 'testProject'
        self.reporter_test_repo = 'https://example.org/repo'
        self.reporter_test_revision = 'd34db33fd43db33f'
        self.reporter_test_branch = "master"
        self.reporter_test_codebase = 'cbgerrit'
        self.reporter_test_change_id = 'I5bdc2e500d00607af53f0fa4df661aada17f81fc'
        self.reporter_test_builder_name = 'Builder0'
        self.reporter_test_props = {
            'Stash_branch': 'refs/changes/34/1234/1',
            'project': self.reporter_test_project,
            'got_revision': self.reporter_test_revision,
            'revision': self.reporter_test_revision,
            'event.change.id': self.reporter_test_change_id,
            'event.change.project': self.reporter_test_project,
            'branch': 'refs/pull/34/merge',
        }
        self.reporter_test_thing_url = 'http://thing.example.com'

    @defer.inlineCallbacks
    def insert_build(self, results, insert_ss=True, parent_plan=False, insert_patch=False):
        yield self.insert_test_data(
            [results],
            results,
            insertSS=insert_ss,
            parentPlan=parent_plan,
            insert_patch=insert_patch,
        )
        build = yield self.master.data.get(("builds", 20))
        return build

    @defer.inlineCallbacks
    def insert_buildset_no_builds(
        self, results, insert_ss=True, parent_plan=False, insert_patch=False
    ):
        yield self.insert_test_data(
            [], results, insertSS=insert_ss, parentPlan=parent_plan, insert_patch=insert_patch
        )
        buildset = yield self.master.data.get(("buildsets", 98))
        return buildset

    @defer.inlineCallbacks
    def insert_build_finished(self, results=SUCCESS, **kwargs):
        return (yield self.insert_build(results=results, **kwargs))

    @defer.inlineCallbacks
    def insert_build_new(self, **kwargs):
        return (yield self.insert_build(results=None, **kwargs))

    @defer.inlineCallbacks
    def insert_buildrequest_new(self, insert_patch=False, **kwargs):
        yield self.master.db.insert_test_data([
            fakedb.Master(id=92),
            fakedb.Worker(id=13, name='wrk'),
            fakedb.Builder(id=79, name='Builder0'),
            fakedb.Builder(id=80, name='Builder1'),
            fakedb.Buildset(id=98, results=None, reason="testReason1", parent_buildid=None),
            fakedb.BuildRequest(id=11, buildsetid=98, builderid=79),
        ])

        patchid = 99 if insert_patch else None

        yield self.master.db.insert_test_data([
            fakedb.BuildsetSourceStamp(buildsetid=98, sourcestampid=234),
            fakedb.SourceStamp(
                id=234,
                branch=self.reporter_test_branch,
                project=self.reporter_test_project,
                revision=self.reporter_test_revision,
                repository=self.reporter_test_repo,
                codebase=self.reporter_test_codebase,
                patchid=patchid,
            ),
            fakedb.Patch(
                id=99,
                patch_base64='aGVsbG8sIHdvcmxk',
                patch_author='him@foo',
                patch_comment='foo',
                subdir='/foo',
                patchlevel=3,
            ),
        ])
        request = yield self.master.data.get(("buildrequests", 11))
        return request

    @defer.inlineCallbacks
    def insert_test_data(
        self, buildResults, finalResult, insertSS=True, parentPlan=False, insert_patch=False
    ):
        rows = [
            fakedb.Master(id=92),
            fakedb.Worker(id=13, name='wrk'),
            fakedb.Builder(id=79, name='Builder0'),
            fakedb.Builder(id=80, name='Builder1'),
            fakedb.Buildset(
                id=98,
                results=finalResult,
                reason="testReason1",
                parent_buildid=19 if parentPlan else None,
            ),
        ]

        if parentPlan:
            rows += [
                fakedb.Worker(id=12, name='wrk_parent'),
                fakedb.Builder(id=78, name='Builder_parent'),
                fakedb.Buildset(id=97, results=finalResult, reason="testReason0"),
                fakedb.BuildRequest(id=10, buildsetid=98, builderid=78),
                fakedb.Build(
                    id=19,
                    number=1,
                    builderid=78,
                    buildrequestid=10,
                    workerid=12,
                    masterid=92,
                    results=finalResult,
                    state_string="buildText",
                ),
            ]

        if insertSS:
            patchid = 99 if insert_patch else None

            rows += [
                fakedb.BuildsetSourceStamp(buildsetid=98, sourcestampid=234),
                fakedb.SourceStamp(
                    id=234,
                    branch=self.reporter_test_branch,
                    project=self.reporter_test_project,
                    revision=self.reporter_test_revision,
                    repository=self.reporter_test_repo,
                    codebase=self.reporter_test_codebase,
                    patchid=patchid,
                ),
                fakedb.Patch(
                    id=99,
                    patch_base64='aGVsbG8sIHdvcmxk',
                    patch_author='him@foo',
                    patch_comment='foo',
                    subdir='/foo',
                    patchlevel=3,
                ),
                fakedb.Change(
                    changeid=13,
                    branch=self.reporter_test_branch,
                    revision='9283',
                    author='me@foo',
                    repository=self.reporter_test_repo,
                    codebase=self.reporter_test_codebase,
                    project='world-domination',
                    sourcestampid=234,
                ),
            ]

        for i, results in enumerate(buildResults):
            started_at = 10000001
            complete_at = None if results is None else 10000005
            rows += [
                fakedb.BuildRequest(id=11 + i, buildsetid=98, builderid=79 + i),
                fakedb.Build(
                    id=20 + i,
                    number=i,
                    builderid=79 + i,
                    buildrequestid=11 + i,
                    workerid=13,
                    masterid=92,
                    results=results,
                    state_string="buildText",
                    started_at=started_at,
                    complete_at=complete_at,
                ),
                fakedb.Step(id=50 + i, buildid=20 + i, number=5, name='make'),
                fakedb.Log(
                    id=60 + i, stepid=50 + i, name='stdio', slug='stdio', type='s', num_lines=7
                ),
                fakedb.LogChunk(
                    logid=60 + i,
                    first_line=0,
                    last_line=1,
                    compressed=0,
                    content='Unicode log with non-ascii (\u00e5\u00e4\u00f6).',
                ),
                fakedb.BuildProperty(buildid=20 + i, name="workername", value="wrk"),
                fakedb.BuildProperty(buildid=20 + i, name="reason", value="because"),
                fakedb.BuildProperty(buildid=20 + i, name="buildername", value="Builder0"),
                fakedb.BuildProperty(buildid=20 + i, name="buildnumber", value=f"{i}"),
                fakedb.BuildProperty(buildid=20 + i, name="scheduler", value="checkin"),
            ]
            for k, v in self.reporter_test_props.items():
                rows += [fakedb.BuildProperty(buildid=20 + i, name=k, value=v)]

        yield self.master.db.insert_test_data(rows)

        self.setup_fake_get_changes_for_build()

    def get_inserted_buildset(self):
        return self.master.data.get(("buildsets", 98))

    def setup_fake_get_changes_for_build(self, has_change=True):
        @defer.inlineCallbacks
        def getChangesForBuild(buildid):
            if not has_change:
                return []

            assert buildid == 20
            ch = yield self.master.db.changes.getChange(13)
            if ch is None:
                return []
            return [ch]

        self.master.db.changes.getChangesForBuild = getChangesForBuild



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import sqlalchemy
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.db import users
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin


class TestUsersConnectorComponent(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True)
        self.db = self.master.db

    # sample user data

    user1_rows = [
        fakedb.User(uid=1, identifier='soap'),
        fakedb.UserInfo(uid=1, attr_type='IPv9', attr_data='0578cc6.8db024'),
    ]

    user2_rows = [
        fakedb.User(uid=2, identifier='lye'),
        fakedb.UserInfo(uid=2, attr_type='git', attr_data='Tyler Durden <tyler@mayhem.net>'),
        fakedb.UserInfo(uid=2, attr_type='irc', attr_data='durden'),
    ]

    user3_rows = [fakedb.User(uid=3, identifier='marla', bb_username='marla', bb_password='cancer')]

    user1_model = users.UserModel(
        uid=1,
        identifier='soap',
        bb_username=None,
        bb_password=None,
        attributes={
            'IPv9': '0578cc6.8db024',
        },
    )

    user2_model = users.UserModel(
        uid=2,
        identifier='lye',
        bb_username=None,
        bb_password=None,
        attributes={
            'irc': 'durden',
            'git': 'Tyler Durden <tyler@mayhem.net>',
        },
    )

    user3_model = users.UserModel(
        uid=3,
        identifier='marla',
        bb_username='marla',
        bb_password='cancer',
        attributes={},
    )

    # tests

    @defer.inlineCallbacks
    def test_addUser_new(self):
        uid = yield self.db.users.findUserByAttr(
            identifier='soap', attr_type='subspace_net_handle', attr_data='Durden0924'
        )

        def thd(conn):
            users_tbl = self.db.model.users
            users_info_tbl = self.db.model.users_info
            users = conn.execute(users_tbl.select()).fetchall()
            infos = conn.execute(users_info_tbl.select()).fetchall()
            self.assertEqual(len(users), 1)
            self.assertEqual(users[0].uid, uid)
            self.assertEqual(users[0].identifier, 'soap')
            self.assertEqual(len(infos), 1)
            self.assertEqual(infos[0].uid, uid)
            self.assertEqual(infos[0].attr_type, 'subspace_net_handle')
            self.assertEqual(infos[0].attr_data, 'Durden0924')

        yield self.db.pool.do(thd)

    @defer.inlineCallbacks
    def test_addUser_existing(self):
        yield self.db.insert_test_data(self.user1_rows)
        uid = yield self.db.users.findUserByAttr(
            identifier='soapy', attr_type='IPv9', attr_data='0578cc6.8db024'
        )

        self.assertEqual(uid, 1)

        def thd(conn):
            users_tbl = self.db.model.users
            users_info_tbl = self.db.model.users_info
            users = conn.execute(users_tbl.select()).fetchall()
            infos = conn.execute(users_info_tbl.select()).fetchall()
            self.assertEqual(len(users), 1)
            self.assertEqual(users[0].uid, uid)
            self.assertEqual(users[0].identifier, 'soap')  # not changed!
            self.assertEqual(len(infos), 1)
            self.assertEqual(infos[0].uid, uid)
            self.assertEqual(infos[0].attr_type, 'IPv9')
            self.assertEqual(infos[0].attr_data, '0578cc6.8db024')

        yield self.db.pool.do(thd)

    @defer.inlineCallbacks
    def test_findUser_existing(self):
        yield self.db.insert_test_data(self.user1_rows + self.user2_rows + self.user3_rows)
        uid = yield self.db.users.findUserByAttr(
            identifier='lye', attr_type='git', attr_data='Tyler Durden <tyler@mayhem.net>'
        )

        self.assertEqual(uid, 2)

        def thd(conn):
            users_tbl = self.db.model.users
            users_info_tbl = self.db.model.users_info
            users = conn.execute(users_tbl.select()).fetchall()
            infos = conn.execute(users_info_tbl.select()).fetchall()
            self.assertEqual(
                (sorted([tuple(u) for u in users]), sorted([tuple(i) for i in infos])),
                (
                    [
                        (1, 'soap', None, None),
                        (2, 'lye', None, None),
                        (3, 'marla', 'marla', 'cancer'),
                    ],
                    [
                        (1, 'IPv9', '0578cc6.8db024'),
                        (2, 'git', 'Tyler Durden <tyler@mayhem.net>'),
                        (2, 'irc', 'durden'),
                    ],
                ),
            )

        yield self.db.pool.do(thd)

    @defer.inlineCallbacks
    def test_addUser_race(self):
        def race_thd(conn):
            # note that this assumes that both inserts can happen "at once".
            # This is the case for DB engines that support transactions, but
            # not for MySQL.  so this test does not detect the potential MySQL
            # failure, which will generally result in a spurious failure.
            conn.execute(self.db.model.users.insert().values(uid=99, identifier='soap'))
            conn.execute(
                self.db.model.users_info.insert().values(
                    uid=99,
                    attr_type='subspace_net_handle',
                    attr_data='Durden0924',
                )
            )
            conn.commit()

        uid = yield self.db.users.findUserByAttr(
            identifier='soap',
            attr_type='subspace_net_handle',
            attr_data='Durden0924',
            _race_hook=race_thd,
        )

        self.assertEqual(uid, 99)

        def thd(conn):
            users_tbl = self.db.model.users
            users_info_tbl = self.db.model.users_info
            users = conn.execute(users_tbl.select()).fetchall()
            infos = conn.execute(users_info_tbl.select()).fetchall()
            self.assertEqual(len(users), 1)
            self.assertEqual(users[0].uid, uid)
            self.assertEqual(users[0].identifier, 'soap')
            self.assertEqual(len(infos), 1)
            self.assertEqual(infos[0].uid, uid)
            self.assertEqual(infos[0].attr_type, 'subspace_net_handle')
            self.assertEqual(infos[0].attr_data, 'Durden0924')

        yield self.db.pool.do(thd)

    @defer.inlineCallbacks
    def test_addUser_existing_identifier(self):
        # see http://trac.buildbot.net/ticket/2587
        yield self.db.insert_test_data(self.user1_rows)
        uid = yield self.db.users.findUserByAttr(
            identifier='soap',  # same identifier
            attr_type='IPv9',
            attr_data='fffffff.ffffff',
        )  # different attr

        # creates a new user
        self.assertNotEqual(uid, 1)

        def thd(conn):
            users_tbl = self.db.model.users
            users_info_tbl = self.db.model.users_info
            users = conn.execute(users_tbl.select().order_by(users_tbl.c.identifier)).fetchall()
            infos = conn.execute(
                users_info_tbl.select().where(users_info_tbl.c.uid == uid)
            ).fetchall()
            self.assertEqual(len(users), 2)
            self.assertEqual(users[1].uid, uid)
            self.assertEqual(users[1].identifier, 'soap_2')  # unique'd
            self.assertEqual(len(infos), 1)
            self.assertEqual(infos[0].attr_type, 'IPv9')
            self.assertEqual(infos[0].attr_data, 'fffffff.ffffff')

        yield self.db.pool.do(thd)

    @defer.inlineCallbacks
    def test_getUser(self):
        yield self.db.insert_test_data(self.user1_rows)

        usdict = yield self.db.users.getUser(1)

        self.assertEqual(usdict, self.user1_model)

    @defer.inlineCallbacks
    def test_getUser_bb(self):
        yield self.db.insert_test_data(self.user3_rows)

        usdict = yield self.db.users.getUser(3)

        self.assertEqual(usdict, self.user3_model)

    @defer.inlineCallbacks
    def test_getUser_multi_attr(self):
        yield self.db.insert_test_data(self.user2_rows)

        usdict = yield self.db.users.getUser(2)

        self.assertEqual(usdict, self.user2_model)

    @defer.inlineCallbacks
    def test_getUser_no_match(self):
        yield self.db.insert_test_data(self.user1_rows)

        none = yield self.db.users.getUser(3)

        self.assertEqual(none, None)

    @defer.inlineCallbacks
    def test_getUsers_none(self):
        res = yield self.db.users.getUsers()

        self.assertEqual(res, [])

    @defer.inlineCallbacks
    def test_getUsers(self):
        yield self.db.insert_test_data(self.user1_rows)

        res = yield self.db.users.getUsers()

        self.assertEqual(res, [users.UserModel(uid=1, identifier='soap')])

    @defer.inlineCallbacks
    def test_getUsers_multiple(self):
        yield self.db.insert_test_data(self.user1_rows + self.user2_rows)

        res = yield self.db.users.getUsers()

        self.assertEqual(
            res,
            [users.UserModel(uid=1, identifier='soap'), users.UserModel(uid=2, identifier='lye')],
        )

    @defer.inlineCallbacks
    def test_getUserByUsername(self):
        yield self.db.insert_test_data(self.user3_rows)

        res = yield self.db.users.getUserByUsername("marla")

        self.assertEqual(res, self.user3_model)

    @defer.inlineCallbacks
    def test_getUserByUsername_no_match(self):
        yield self.db.insert_test_data(self.user3_rows)

        none = yield self.db.users.getUserByUsername("tyler")

        self.assertEqual(none, None)

    @defer.inlineCallbacks
    def test_updateUser_existing_type(self):
        yield self.db.insert_test_data(self.user1_rows)

        yield self.db.users.updateUser(uid=1, attr_type='IPv9', attr_data='abcd.1234')

        usdict = yield self.db.users.getUser(1)

        self.assertEqual(usdict.attributes['IPv9'], 'abcd.1234')
        self.assertEqual(usdict.identifier, 'soap')  # no change

    @defer.inlineCallbacks
    def test_updateUser_new_type(self):
        yield self.db.insert_test_data(self.user1_rows)

        yield self.db.users.updateUser(uid=1, attr_type='IPv4', attr_data='123.134.156.167')

        usdict = yield self.db.users.getUser(1)

        self.assertEqual(usdict.attributes['IPv4'], '123.134.156.167')
        self.assertEqual(usdict.attributes['IPv9'], '0578cc6.8db024')  # no change
        self.assertEqual(usdict.identifier, 'soap')  # no change

    @defer.inlineCallbacks
    def test_updateUser_identifier(self):
        yield self.db.insert_test_data(self.user1_rows)

        yield self.db.users.updateUser(uid=1, identifier='lye')

        usdict = yield self.db.users.getUser(1)

        self.assertEqual(usdict.identifier, 'lye')
        self.assertEqual(usdict.attributes['IPv9'], '0578cc6.8db024')  # no change

    @defer.inlineCallbacks
    def test_updateUser_bb(self):
        yield self.db.insert_test_data(self.user3_rows)

        yield self.db.users.updateUser(uid=3, bb_username='boss', bb_password='fired')

        usdict = yield self.db.users.getUser(3)

        self.assertEqual(usdict.bb_username, 'boss')
        self.assertEqual(usdict.bb_password, 'fired')
        self.assertEqual(usdict.identifier, 'marla')  # no change

    @defer.inlineCallbacks
    def test_updateUser_all(self):
        yield self.db.insert_test_data(self.user1_rows)

        yield self.db.users.updateUser(
            uid=1,
            identifier='lye',
            bb_username='marla',
            bb_password='cancer',
            attr_type='IPv4',
            attr_data='123.134.156.167',
        )

        usdict = yield self.db.users.getUser(1)

        self.assertEqual(
            usdict,
            users.UserModel(
                uid=1,
                identifier='lye',
                bb_username='marla',
                bb_password='cancer',
                attributes={
                    'IPv4': '123.134.156.167',
                    'IPv9': '0578cc6.8db024',  # no change
                },
            ),
        )

    @defer.inlineCallbacks
    def test_updateUser_race(self):
        # called from the db thread, this opens a *new* connection (to avoid
        # the existing transaction) and executes a conflicting insert in that
        # connection.  This will cause the insert in the db method to fail, and
        # the data in this insert (8.8.8.8) will appear below.
        race_condition_committed = []
        if (
            self.db.pool.engine.dialect.name == 'sqlite'
            and self.db.pool.engine.url.database not in [None, ':memory:']
        ):
            # It's not easy to work with file-based SQLite via multiple
            # connections, because SQLAlchemy (in it's default configuration)
            # locks file during working session.
            # TODO: This probably can be supported.
            raise unittest.SkipTest("It's hard to test race condition with not in-memory SQLite")

        def race_thd(conn):
            conn = self.db.pool.engine.connect()
            try:
                r = conn.execute(
                    self.db.model.users_info.insert().values(
                        uid=1, attr_type='IPv4', attr_data='8.8.8.8'
                    )
                )
                conn.commit()
                r.close()
                conn.close()
                race_condition_committed.append(True)
            except (sqlalchemy.exc.IntegrityError, sqlalchemy.exc.ProgrammingError):
                # some engine (mysql innodb) will enforce lock until the transaction is over
                race_condition_committed.append(False)
                # scope variable, we modify a list so that modification is visible in parent scope

        yield self.db.insert_test_data(self.user1_rows)

        yield self.db.users.updateUser(
            uid=1, attr_type='IPv4', attr_data='123.134.156.167', _race_hook=race_thd
        )

        if not race_condition_committed:
            raise RuntimeError('programmer error: race condition was not called')

        usdict = yield self.db.users.getUser(1)

        self.assertEqual(usdict.identifier, 'soap')
        if race_condition_committed[0] == self.db.has_native_upsert:
            self.assertEqual(usdict.attributes['IPv4'], '123.134.156.167')
        else:
            self.assertEqual(usdict.attributes['IPv4'], '8.8.8.8')
        self.assertEqual(usdict.attributes['IPv9'], '0578cc6.8db024')  # no change

    @defer.inlineCallbacks
    def test_update_NoMatch_identifier(self):
        yield self.db.insert_test_data(self.user1_rows)

        yield self.db.users.updateUser(uid=3, identifier='abcd')

        usdict = yield self.db.users.getUser(1)

        self.assertEqual(usdict.identifier, 'soap')  # no change

    @defer.inlineCallbacks
    def test_update_NoMatch_attribute(self):
        yield self.db.insert_test_data(self.user1_rows)

        yield self.db.users.updateUser(uid=3, attr_type='abcd', attr_data='efgh')

        usdict = yield self.db.users.getUser(1)

        self.assertEqual(usdict.attributes['IPv9'], '0578cc6.8db024')  # no change

    @defer.inlineCallbacks
    def test_update_NoMatch_bb(self):
        yield self.db.insert_test_data(self.user1_rows)

        yield self.db.users.updateUser(uid=3, attr_type='marla', attr_data='cancer')

        usdict = yield self.db.users.getUser(1)

        self.assertEqual(usdict.attributes['IPv9'], '0578cc6.8db024')  # no change

    @defer.inlineCallbacks
    def test_removeUser_uid(self):
        yield self.db.insert_test_data(self.user1_rows)

        yield self.db.users.removeUser(1)

        def thd(conn):
            r = conn.execute(self.db.model.users.select())
            r = r.fetchall()
            self.assertEqual(len(r), 0)

        yield self.db.pool.do(thd)

    @defer.inlineCallbacks
    def test_removeNoMatch(self):
        yield self.db.insert_test_data(self.user1_rows)

        yield self.db.users.removeUser(uid=3)

    @defer.inlineCallbacks
    def test_identifierToUid_NoMatch(self):
        res = yield self.db.users.identifierToUid(identifier="soap")

        self.assertEqual(res, None)

    @defer.inlineCallbacks
    def test_identifierToUid_match(self):
        yield self.db.insert_test_data(self.user1_rows)

        res = yield self.db.users.identifierToUid(identifier="soap")

        self.assertEqual(res, 1)



# This file is part of Buildbot. Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest import mock

from twisted.internet import defer
from twisted.internet.address import IPv4Address
from twisted.spread import pb as twisted_pb
from twisted.trial import unittest

from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import protocols as util_protocols
from buildbot.worker.protocols import base
from buildbot.worker.protocols import pb


class TestListener(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self)

    def makeListener(self):
        listener = pb.Listener(self.master)
        return listener

    def test_constructor(self):
        listener = pb.Listener(self.master)
        self.assertEqual(listener.master, self.master)
        self.assertEqual(listener._registrations, {})

    @defer.inlineCallbacks
    def test_updateRegistration_simple(self):
        listener = pb.Listener(self.master)
        reg = yield listener.updateRegistration('example', 'pass', 'tcp:1234')
        self.assertEqual(self.master.pbmanager._registrations, [('tcp:1234', 'example', 'pass')])
        self.assertEqual(listener._registrations['example'], ('pass', 'tcp:1234', reg))

    @defer.inlineCallbacks
    def test_updateRegistration_pass_changed(self):
        listener = pb.Listener(self.master)
        listener.updateRegistration('example', 'pass', 'tcp:1234')
        reg1 = yield listener.updateRegistration('example', 'pass1', 'tcp:1234')
        self.assertEqual(listener._registrations['example'], ('pass1', 'tcp:1234', reg1))
        self.assertEqual(self.master.pbmanager._unregistrations, [('tcp:1234', 'example')])

    @defer.inlineCallbacks
    def test_updateRegistration_port_changed(self):
        listener = pb.Listener(self.master)
        listener.updateRegistration('example', 'pass', 'tcp:1234')
        reg1 = yield listener.updateRegistration('example', 'pass', 'tcp:4321')
        self.assertEqual(listener._registrations['example'], ('pass', 'tcp:4321', reg1))
        self.assertEqual(self.master.pbmanager._unregistrations, [('tcp:1234', 'example')])

    @defer.inlineCallbacks
    def test_create_connection(self):
        listener = pb.Listener(self.master)
        worker = mock.Mock()
        worker.workername = 'test'
        mind = mock.Mock()

        listener.updateRegistration('example', 'pass', 'tcp:1234')
        self.master.workers.register(worker)
        conn = yield listener._create_connection(mind, worker.workername)

        mind.broker.transport.setTcpKeepAlive.assert_called_with(1)
        self.assertIsInstance(conn, pb.Connection)


class TestConnectionApi(
    util_protocols.ConnectionInterfaceTest, TestReactorMixin, unittest.TestCase
):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self)
        self.conn = pb.Connection(self.master, mock.Mock(), mock.Mock())


class TestConnection(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self)
        self.mind = mock.Mock()
        self.worker = mock.Mock()

    def test_constructor(self):
        conn = pb.Connection(self.master, self.worker, self.mind)

        self.assertEqual(conn.mind, self.mind)
        self.assertEqual(conn.master, self.master)
        self.assertEqual(conn.worker, self.worker)

    @defer.inlineCallbacks
    def test_attached(self):
        conn = pb.Connection(self.master, self.worker, self.mind)
        att = yield conn.attached(self.mind)

        self.worker.attached.assert_called_with(conn)
        self.assertEqual(att, conn)

        self.reactor.pump([10] * 361)
        expected_call = [
            mock.call('print', message="keepalive"),
        ]
        self.assertEqual(self.mind.callRemote.call_args_list, expected_call)

        conn.detached(self.mind)
        yield conn.waitShutdown()

    @defer.inlineCallbacks
    def test_detached(self):
        conn = pb.Connection(self.master, self.worker, self.mind)
        conn.attached(self.mind)
        conn.detached(self.mind)

        self.assertEqual(conn.keepalive_timer, None)
        self.assertEqual(conn.mind, None)
        yield conn.waitShutdown()

    def test_loseConnection(self):
        conn = pb.Connection(self.master, self.worker, self.mind)
        conn.loseConnection()

        self.assertEqual(conn.keepalive_timer, None)
        conn.mind.broker.transport.loseConnection.assert_called_with()

    def test_remotePrint(self):
        conn = pb.Connection(self.master, self.worker, self.mind)
        conn.remotePrint(message='test')
        conn.mind.callRemote.assert_called_with('print', message='test')

    @defer.inlineCallbacks
    def test_remoteGetWorkerInfo_slave(self):
        def side_effect(*args, **kwargs):
            if args[0] == 'getWorkerInfo':
                return defer.fail(
                    twisted_pb.RemoteError('twisted.spread.flavors.NoSuchMethod', None, None)
                )
            if args[0] == 'getSlaveInfo':
                return defer.succeed({'info': 'test'})
            if args[0] == 'getCommands':
                return defer.succeed({'x': 1, 'y': 2})
            if args[0] == 'getVersion':
                return defer.succeed('TheVersion')
            return None

        self.mind.callRemote.side_effect = side_effect
        conn = pb.Connection(self.master, self.worker, self.mind)
        info = yield conn.remoteGetWorkerInfo()

        r = {'info': 'test', 'worker_commands': {'y': 2, 'x': 1}, 'version': 'TheVersion'}
        self.assertEqual(info, r)
        expected_calls = [
            mock.call('getWorkerInfo'),
            mock.call(
                'print',
                message='buildbot-slave detected, failing back to deprecated buildslave API. '
                '(Ignoring missing getWorkerInfo method.)',
            ),
            mock.call('getSlaveInfo'),
            mock.call('getCommands'),
            mock.call('getVersion'),
        ]
        self.assertEqual(self.mind.callRemote.call_args_list, expected_calls)

    @defer.inlineCallbacks
    def test_remoteGetWorkerInfo_slave_2_16(self):
        """In buildslave 2.16 all information about worker is retrieved in
        a single getSlaveInfo() call."""

        def side_effect(*args, **kwargs):
            if args[0] == 'getWorkerInfo':
                return defer.fail(
                    twisted_pb.RemoteError('twisted.spread.flavors.NoSuchMethod', None, None)
                )
            if args[0] == 'getSlaveInfo':
                return defer.succeed({
                    'info': 'test',
                    'slave_commands': {'x': 1, 'y': 2},
                    'version': 'TheVersion',
                })
            if args[0] == 'print':
                return None
            raise ValueError(f"Command unknown: {args}")

        self.mind.callRemote.side_effect = side_effect
        conn = pb.Connection(self.master, self.worker, self.mind)
        info = yield conn.remoteGetWorkerInfo()

        r = {'info': 'test', 'worker_commands': {'y': 2, 'x': 1}, 'version': 'TheVersion'}
        self.assertEqual(info, r)
        expected_calls = [
            mock.call('getWorkerInfo'),
            mock.call(
                'print',
                message='buildbot-slave detected, failing back to deprecated buildslave API. '
                '(Ignoring missing getWorkerInfo method.)',
            ),
            mock.call('getSlaveInfo'),
        ]
        self.assertEqual(self.mind.callRemote.call_args_list, expected_calls)

    @defer.inlineCallbacks
    def test_remoteGetWorkerInfo_worker(self):
        def side_effect(*args, **kwargs):
            if args[0] == 'getWorkerInfo':
                return defer.succeed({
                    'info': 'test',
                    'worker_commands': {'y': 2, 'x': 1},
                    'version': 'TheVersion',
                })
            raise ValueError(f"Command unknown: {args}")

        self.mind.callRemote.side_effect = side_effect
        conn = pb.Connection(self.master, self.worker, self.mind)
        info = yield conn.remoteGetWorkerInfo()

        r = {'info': 'test', 'worker_commands': {'y': 2, 'x': 1}, 'version': 'TheVersion'}
        self.assertEqual(info, r)
        expected_calls = [mock.call('getWorkerInfo')]
        self.assertEqual(self.mind.callRemote.call_args_list, expected_calls)

    @defer.inlineCallbacks
    def test_remoteGetWorkerInfo_getWorkerInfo_fails(self):
        def side_effect(*args, **kwargs):
            if args[0] == 'getWorkerInfo':
                return defer.fail(
                    twisted_pb.RemoteError('twisted.spread.flavors.NoSuchMethod', None, None)
                )
            if args[0] == 'getSlaveInfo':
                return defer.fail(
                    twisted_pb.RemoteError('twisted.spread.flavors.NoSuchMethod', None, None)
                )
            if args[0] == 'getCommands':
                return defer.succeed({'x': 1, 'y': 2})
            if args[0] == 'getVersion':
                return defer.succeed('TheVersion')
            if args[0] == 'print':
                return None
            raise ValueError(f"Command unknown: {args}")

        self.mind.callRemote.side_effect = side_effect
        conn = pb.Connection(self.master, self.worker, self.mind)
        info = yield conn.remoteGetWorkerInfo()

        r = {'worker_commands': {'y': 2, 'x': 1}, 'version': 'TheVersion'}
        self.assertEqual(info, r)
        expected_calls = [
            mock.call('getWorkerInfo'),
            mock.call(
                'print',
                message='buildbot-slave detected, failing back to deprecated buildslave API. '
                '(Ignoring missing getWorkerInfo method.)',
            ),
            mock.call('getSlaveInfo'),
            mock.call('getCommands'),
            mock.call('getVersion'),
        ]
        self.assertEqual(self.mind.callRemote.call_args_list, expected_calls)

    @defer.inlineCallbacks
    def test_remoteGetWorkerInfo_no_info(self):
        # All remote commands tried in remoteGetWorkerInfo are unavailable.
        # This should be real old worker...
        def side_effect(*args, **kwargs):
            if args[0] == 'print':
                return None
            return defer.fail(
                twisted_pb.RemoteError('twisted.spread.flavors.NoSuchMethod', None, None)
            )

        self.mind.callRemote.side_effect = side_effect
        conn = pb.Connection(self.master, self.worker, self.mind)
        info = yield conn.remoteGetWorkerInfo()

        r = {}
        self.assertEqual(info, r)
        expected_calls = [
            mock.call('getWorkerInfo'),
            mock.call(
                'print',
                message='buildbot-slave detected, failing back to deprecated buildslave API. '
                '(Ignoring missing getWorkerInfo method.)',
            ),
            mock.call('getSlaveInfo'),
            mock.call('getCommands'),
            mock.call('getVersion'),
        ]
        self.assertEqual(self.mind.callRemote.call_args_list, expected_calls)

    @defer.inlineCallbacks
    def test_remoteSetBuilderList(self):
        builders = ['builder1', 'builder2']
        self.mind.callRemote.return_value = defer.succeed(builders)
        conn = pb.Connection(self.master, self.worker, self.mind)
        r = yield conn.remoteSetBuilderList(builders)

        self.assertEqual(r, builders)
        self.assertEqual(conn.builders, builders)
        self.mind.callRemote.assert_called_with('setBuilderList', builders)

    def test_remoteStartCommand(self):
        builders = ['builder']
        ret_val = {'builder': mock.Mock()}
        self.mind.callRemote.return_value = defer.succeed(ret_val)
        conn = pb.Connection(self.master, self.worker, self.mind)
        conn.remoteSetBuilderList(builders)

        RCInstance = base.RemoteCommandImpl()
        builder_name = "builder"
        commandID = None
        remote_command = "command"
        args = {"args": 'args'}

        conn.remoteStartCommand(RCInstance, builder_name, commandID, remote_command, args)

        callargs = ret_val['builder'].callRemote.call_args_list[0][0]
        callargs_without_rc = (callargs[0], callargs[2], callargs[3], callargs[4])
        self.assertEqual(callargs_without_rc, ('startCommand', commandID, remote_command, args))
        self.assertIsInstance(callargs[1], pb.RemoteCommand)
        self.assertEqual(callargs[1].impl, RCInstance)

    @defer.inlineCallbacks
    def test_do_keepalive(self):
        conn = pb.Connection(self.master, self.worker, self.mind)
        yield conn._do_keepalive()

        self.mind.callRemote.assert_called_with('print', message="keepalive")

    def test_remoteShutdown(self):
        self.mind.callRemote.return_value = defer.succeed(None)
        conn = pb.Connection(self.master, self.worker, self.mind)
        # note that we do not test the "old way", as it is now *very* old.
        conn.remoteShutdown()

        self.mind.callRemote.assert_called_with('shutdown')

    def test_remoteStartBuild(self):
        conn = pb.Connection(self.master, self.worker, self.mind)
        builders = {'builder': mock.Mock()}
        self.mind.callRemote.return_value = defer.succeed(builders)
        conn = pb.Connection(self.master, self.worker, self.mind)
        conn.remoteSetBuilderList(builders)

        conn.remoteStartBuild('builder')

        builders['builder'].callRemote.assert_called_with('startBuild')

    @defer.inlineCallbacks
    def test_startStopKeepaliveTimer(self):
        conn = pb.Connection(self.master, self.worker, self.mind)
        conn.startKeepaliveTimer()

        self.mind.callRemote.assert_not_called()
        self.reactor.pump([10] * 361)
        expected_call = [
            mock.call('print', message="keepalive"),
        ]
        self.assertEqual(self.mind.callRemote.call_args_list, expected_call)

        self.reactor.pump([10] * 361)
        expected_calls = [
            mock.call('print', message="keepalive"),
            mock.call('print', message="keepalive"),
        ]
        self.assertEqual(self.mind.callRemote.call_args_list, expected_calls)

        conn.stopKeepaliveTimer()
        yield conn.waitShutdown()

    def test_perspective_shutdown(self):
        conn = pb.Connection(self.master, self.worker, self.mind)
        conn.perspective_shutdown()

        conn.worker.shutdownRequested.assert_called_with()
        conn.worker.messageReceivedFromWorker.assert_called_with()

    def test_perspective_keepalive(self):
        conn = pb.Connection(self.master, self.worker, self.mind)
        conn.perspective_keepalive()

        conn.worker.messageReceivedFromWorker.assert_called_with()

    def test_get_peer(self):
        conn = pb.Connection(self.master, self.worker, self.mind)
        conn.mind.broker.transport.getPeer.return_value = IPv4Address(
            "TCP",
            "ip",
            "port",
        )
        self.assertEqual(conn.get_peer(), "ip:port")


class Test_wrapRemoteException(unittest.TestCase):
    def test_raises_NoSuchMethod(self):
        def f():
            with pb._wrapRemoteException():
                raise twisted_pb.RemoteError('twisted.spread.flavors.NoSuchMethod', None, None)

        with self.assertRaises(pb._NoSuchMethod):
            f()

    def test_raises_unknown(self):
        class Error(Exception):
            pass

        def f():
            with pb._wrapRemoteException():
                raise Error()

        with self.assertRaises(Error):
            f()

    def test_raises_RemoteError(self):
        def f():
            with pb._wrapRemoteException():
                raise twisted_pb.RemoteError('twisted.spread.flavors.ProtocolError', None, None)

        with self.assertRaises(twisted_pb.RemoteError):
            f()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

from twisted.internet import defer
from twisted.python import log
from twisted.trial import unittest

from buildbot.util import eventual


class Eventually(unittest.TestCase):
    def setUp(self):
        # reset the queue to its base state
        eventual._theSimpleQueue = eventual._SimpleCallQueue()
        self.old_log_err = log.err
        self.results = []

    def tearDown(self):
        log.err = self.old_log_err
        return eventual.flushEventualQueue()

    # utility callback
    def cb(self, *args, **kwargs):
        r = args
        if kwargs:
            r = (*r, kwargs)
        self.results.append(r)

    # flush the queue and assert results
    @defer.inlineCallbacks
    def assertResults(self, exp):
        yield eventual.flushEventualQueue()

        self.assertEqual(self.results, exp)

    # tests

    def test_eventually_calls(self):
        eventual.eventually(self.cb)
        return self.assertResults([()])

    def test_eventually_args(self):
        eventual.eventually(self.cb, 1, 2, a='a')
        return self.assertResults([(1, 2, {"a": 'a'})])

    def test_eventually_err(self):
        # monkey-patch log.err; this is restored by tearDown
        def cb_err():
            self.results.append("err")

        log.err = cb_err

        def cb_fails():
            raise RuntimeError("should not cause test failure")

        eventual.eventually(cb_fails)
        return self.assertResults(['err'])

    def test_eventually_butNotNow(self):
        eventual.eventually(self.cb, 1)
        self.assertFalse(self.results)
        return self.assertResults([(1,)])

    def test_eventually_order(self):
        eventual.eventually(self.cb, 1)
        eventual.eventually(self.cb, 2)
        eventual.eventually(self.cb, 3)
        return self.assertResults([(1,), (2,), (3,)])

    def test_flush_waitForChainedEventuallies(self):
        def chain(n):
            self.results.append(n)
            if n <= 0:
                return
            eventual.eventually(chain, n - 1)

        chain(3)
        # (the flush this tests is implicit in assertResults)
        return self.assertResults([3, 2, 1, 0])

    def test_flush_waitForTreeEventuallies(self):
        # a more complex set of eventualities
        def tree(n):
            self.results.append(n)
            if n <= 0:
                return
            eventual.eventually(tree, n - 1)
            eventual.eventually(tree, n - 1)

        tree(2)
        # (the flush this tests is implicit in assertResults)
        return self.assertResults([2, 1, 1, 0, 0, 0, 0])

    def test_flush_duringTurn(self):
        testd = defer.Deferred()

        def cb():
            d = eventual.flushEventualQueue()
            d.addCallback(testd.callback)

        eventual.eventually(cb)
        return testd

    def test_fireEventually_call(self):
        d = eventual.fireEventually(13)
        d.addCallback(self.cb)
        return self.assertResults([(13,)])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import json
import zlib
from typing import TYPE_CHECKING
from unittest import mock

from twisted.internet import defer
from twisted.internet import protocol
from twisted.internet import reactor
from twisted.trial import unittest
from twisted.web import client
from twisted.web.http_headers import Headers

from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.util import www
from buildbot.util import bytes2unicode
from buildbot.util import unicode2bytes
from buildbot.util.twisted import async_to_deferred
from buildbot.www import auth
from buildbot.www import authz
from buildbot.www import service as wwwservice

if TYPE_CHECKING:
    from typing import Callable

SOMETIME = 1348971992
OTHERTIME = 1008971992


class BodyReader(protocol.Protocol):
    # an IProtocol that reads the entire HTTP body and then calls back
    # with it

    def __init__(self, finishedDeferred):
        self.body = []
        self.finishedDeferred = finishedDeferred

    def dataReceived(self, data):
        self.body.append(data)

    def connectionLost(self, reason):
        if reason.check(client.ResponseDone):
            self.finishedDeferred.callback(b''.join(self.body))
        else:
            self.finishedDeferred.errback(reason)


class Www(www.RequiresWwwMixin, unittest.TestCase):
    master = None

    @defer.inlineCallbacks
    def setUp(self):
        # set up a full master serving HTTP
        master = yield fakemaster.make_master(
            self,
            wantRealReactor=True,
            wantDb=True,
            wantData=True,
            sqlite_memory=False,
            auto_shutdown=False,
        )

        master.config.www = {
            "port": 'tcp:0:interface=127.0.0.1',
            "debug": True,
            "auth": auth.NoAuth(),
            "authz": authz.Authz(),
            "avatar_methods": [],
            "logfileName": 'http.log',
        }
        master.www = wwwservice.WWWService()
        yield master.www.setServiceParent(master)
        yield master.www.startService()
        yield master.www.reconfigServiceWithBuildbotConfig(master.config)
        session = mock.Mock()
        session.uid = "0"
        master.www.site.sessionFactory = mock.Mock(return_value=session)

        # now that we have a port, construct the real URL and insert it into
        # the config.  The second reconfig isn't really required, but doesn't
        # hurt.
        self.url = f'http://127.0.0.1:{master.www.getPortnum()}/'
        self.url = unicode2bytes(self.url)
        master.config.buildbotURL = self.url
        yield master.www.reconfigServiceWithBuildbotConfig(master.config)

        self.master = master

        self.addCleanup(self.master.test_shutdown)
        self.addCleanup(self.master.www.stopService)

        # build an HTTP agent, using an explicit connection pool if Twisted
        # supports it (Twisted 13.0.0 and up)
        if hasattr(client, 'HTTPConnectionPool'):
            self.pool = client.HTTPConnectionPool(reactor)
            self.agent = client.Agent(reactor, pool=self.pool)
            self.addCleanup(self.pool.closeCachedConnections)
        else:
            self.pool = None
            self.agent = client.Agent(reactor)

    @defer.inlineCallbacks
    def apiGet(self, url, expect200=True):
        pg = yield self.agent.request(b'GET', url)

        # this is kind of obscene, but protocols are like that
        d = defer.Deferred()
        bodyReader = BodyReader(d)
        pg.deliverBody(bodyReader)
        body = yield d

        # check this *after* reading the body, otherwise Trial will
        # complain that the response is half-read
        if expect200 and pg.code != 200:
            self.fail(f"did not get 200 response for '{url}'")

        return json.loads(bytes2unicode(body))

    def link(self, suffix):
        return self.url + b'api/v2/' + suffix

    # tests

    # There's no need to be exhaustive here.  The intent is to test that data
    # can get all the way from the DB to a real HTTP client, and a few
    # resources will be sufficient to demonstrate that.

    @defer.inlineCallbacks
    def test_masters(self):
        yield self.master.db.insert_test_data([
            fakedb.Master(id=7, active=0, last_active=SOMETIME),
            fakedb.Master(id=8, active=1, last_active=OTHERTIME),
        ])

        res = yield self.apiGet(self.link(b'masters'))
        self.assertEqual(
            res,
            {
                'masters': [
                    {
                        'active': False,
                        'masterid': 7,
                        'name': 'master-7',
                        'last_active': SOMETIME,
                    },
                    {
                        'active': True,
                        'masterid': 8,
                        'name': 'master-8',
                        'last_active': OTHERTIME,
                    },
                ],
                'meta': {
                    'total': 2,
                },
            },
        )

        res = yield self.apiGet(self.link(b'masters/7'))
        self.assertEqual(
            res,
            {
                'masters': [
                    {
                        'active': False,
                        'masterid': 7,
                        'name': 'master-7',
                        'last_active': SOMETIME,
                    },
                ],
                'meta': {},
            },
        )

    async def _test_compression(
        self,
        encoding: bytes,
        decompress_fn: Callable[[bytes], bytes],
    ) -> None:
        assert self.master
        await self.master.db.insert_test_data([
            fakedb.Master(id=7, active=0, last_active=SOMETIME),
        ])

        pg = await self.agent.request(
            b'GET',
            self.link(b'masters/7'),
            headers=Headers({b'accept-encoding': [encoding]}),
        )

        # this is kind of obscene, but protocols are like that
        d: defer.Deferred[bytes] = defer.Deferred()
        bodyReader = BodyReader(d)
        pg.deliverBody(bodyReader)
        body = await d

        self.assertEqual(pg.headers.getRawHeaders(b'content-encoding'), [encoding])

        response = json.loads(bytes2unicode(decompress_fn(body)))
        self.assertEqual(
            response,
            {
                'masters': [
                    {
                        'active': False,
                        'masterid': 7,
                        'name': 'master-7',
                        'last_active': SOMETIME,
                    },
                ],
                'meta': {},
            },
        )

    @async_to_deferred
    async def test_gzip_compression(self):
        await self._test_compression(
            b'gzip',
            decompress_fn=lambda body: zlib.decompress(
                body,
                # use largest wbits possible as twisted customize it
                # see: https://docs.python.org/3/library/zlib.html#zlib.decompress
                wbits=47,
            ),
        )

    @async_to_deferred
    async def test_brotli_compression(self):
        try:
            import brotli
        except ImportError as e:
            raise unittest.SkipTest("brotli not installed, skip the test") from e
        await self._test_compression(b'br', decompress_fn=brotli.decompress)

    @async_to_deferred
    async def test_zstandard_compression(self):
        try:
            import zstandard
        except ImportError as e:
            raise unittest.SkipTest("zstandard not installed, skip the test") from e

        def _decompress(data):
            # zstd cannot decompress data compressed with stream api with a non stream api
            decompressor = zstandard.ZstdDecompressor()
            decompressobj = decompressor.decompressobj()
            return decompressobj.decompress(data) + decompressobj.flush()

        await self._test_compression(b'zstd', decompress_fn=_decompress)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
# Copyright 2013 (c) Mamba Team

from __future__ import annotations

import json
from typing import TYPE_CHECKING
from typing import Any

from dateutil.parser import parse as dateparse
from twisted.internet import defer
from twisted.python import log

from buildbot.util import bytes2unicode
from buildbot.www.hooks.base import BaseHookHandler

if TYPE_CHECKING:
    from twisted.web.server import Request

_HEADER_EVENT = b'X-Event-Key'


class BitBucketHandler(BaseHookHandler):
    def getChanges(
        self, request: Request
    ) -> defer.Deferred[tuple[list[dict[str, Any]], str | None]]:
        """Catch a POST request from BitBucket and start a build process

        Check the URL below if you require more information about payload
        https://confluence.atlassian.com/display/BITBUCKET/POST+Service+Management

        :param request: the http request Twisted object
        :param options: additional options
        """

        assert request.args is not None

        event_type = bytes2unicode(request.getHeader(_HEADER_EVENT))
        payload = json.loads(bytes2unicode(request.args[b'payload'][0]))
        repo_url = f"{payload['canon_url']}{payload['repository']['absolute_url']}"
        project = bytes2unicode(request.args.get(b'project', [b''])[0])

        changes = []
        for commit in payload['commits']:
            changes.append({
                'author': commit['raw_author'],
                'files': [f['file'] for f in commit['files']],
                'comments': commit['message'],
                'revision': commit['raw_node'],
                'when_timestamp': dateparse(commit['utctimestamp']),
                'branch': commit['branch'],
                'revlink': f"{repo_url}commits/{commit['raw_node']}",
                'repository': repo_url,
                'project': project,
                'properties': {
                    'event': event_type,
                },
            })
            log.msg(f"New revision: {commit['node']}")

        log.msg(f'Received {len(changes)} changes from bitbucket')
        return defer.succeed((changes, payload['repository']['scm']))


bitbucket = BitBucketHandler



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import re
from datetime import datetime

from twisted.internet import defer
from twisted.trial import unittest
from twisted.web.error import Error

from buildbot.changes.bitbucket import BitbucketPullrequestPoller
from buildbot.test.fake import httpclientservice as fakehttpclientservice
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import changesource
from buildbot.test.util.logging import LoggingMixin


class SourceRest:
    """https://api.bitbucket.org/2.0/repositories/{owner}/{slug}"""

    template = """\
{
    "hash": "%(hash)s",
    "links": {
        "html": {
            "href": "https://bitbucket.org/%(owner)s/%(slug)s/commits/%(short_hash)s"
        },
        "diff": {
            "href": "https://api.bitbucket.org/2.0/repositories/%(owner)s/%(slug)s/diff/%(hash)s"
        }
    },
    "repository": {
        "links": {
            "self": {
                "href": "https://api.bitbucket.org/2.0/repositories/%(owner)s/%(slug)s"
            }
        }
    },
    "date": "%(date)s"

}
"""
    repo_template = """\
{
    "links": {
        "html": {
            "href": "https://bitbucket.org/%(owner)s/%(slug)s"
        }
    }
}
"""

    def __init__(self, owner, slug, hash, date):
        self.owner = owner
        self.slug = slug
        self.hash = hash
        self.date = date

    def response(self):
        return self.template % {
            "owner": self.owner,
            "slug": self.slug,
            "hash": self.hash,
            "short_hash": self.hash[0:12],
            "date": self.date,
        }

    def repo_response(self):
        return self.repo_template % {
            "owner": self.owner,
            "slug": self.slug,
        }

    def diff_response(self):
        return """
diff --git a/path/to/a/file.txt b/path/to/a/file.txt
index 3e59caa..be38dcf 100644
--- a/path/to/a/file.txt
+++ b/path/to/a/file.txt
@@ -1 +1 @@
-// header
+// Header
"""


class PullRequestRest:
    """https://api.bitbucket.org/2.0/repositories/{owner}/{slug}/pullrequests/{pull_request_id}"""

    template = """\
{

    "description": "%(description)s",
    "title": "%(title)s",
    "source": {
        "commit": {
            "hash": "%(hash)s",
            "links": {
                "self": {
                    "href": "https://api.bitbucket.org/2.0/repositories/%(owner)s/%(slug)s/commit/%(hash)s"
                }
            }
        }
    },
    "state": "OPEN",
    "author": {
        "display_name": "%(display_name)s"
    },
    "created_on": "%(created_on)s",
    "participants": [
    ],
    "updated_on": "%(updated_on)s",
    "merge_commit": null,
    "id": %(id)d

}
"""

    def __init__(self, nr, title, description, display_name, source, created_on, updated_on=None):
        self.nr = nr
        self.title = title
        self.description = description
        self.display_name = display_name
        self.source = source
        self.created_on = created_on
        if updated_on:
            self.updated_on = updated_on
        else:
            self.updated_on = self.created_on

    def response(self):
        return self.template % {
            "description": self.description,
            "title": self.title,
            "hash": self.source.hash,
            "short_hash": self.source.hash[0:12],
            "owner": self.source.owner,
            "slug": self.source.slug,
            "display_name": self.display_name,
            "created_on": self.created_on,
            "updated_on": self.updated_on,
            "id": self.nr,
        }


class PullRequestListRest:
    """https://api.bitbucket.org/2.0/repositories/{owner}/{slug}/pullrequests"""

    template = """\
        {
            "description": "%(description)s",
            "links": {
                "self": {
                    "href": "https://api.bitbucket.org/2.0/repositories/%(owner)s/%(slug)s/pullrequests/%(id)d"
                },
                "html": {
                    "href": "https://bitbucket.org/%(owner)s/%(slug)s/pull-request/%(id)d"
                }
            },
            "author": {
                "display_name": "%(display_name)s"
            },
            "title": "%(title)s",
            "source": {
                "commit": {
                    "hash": "%(short_hash)s",
                    "links": {
                        "self": {
                            "href": "https://api.bitbucket.org/2.0/repositories/%(src_owner)s/%(src_slug)s/commit/%(short_hash)s"
                        }
                    }
                },
                "repository": {
                    "links": {
                        "self": {
                            "href": "https://api.bitbucket.org/2.0/repositories/%(src_owner)s/%(src_slug)s"
                        }
                    }
                },
                "branch": {
                    "name": "default"
                }
            },
            "state": "OPEN",
            "created_on": "%(created_on)s",
            "updated_on": "%(updated_on)s",
            "merge_commit": null,
            "id": %(id)s
        }
"""

    def __init__(self, owner, slug, prs):
        self.owner = owner
        self.slug = slug
        self.prs = prs

        self.pr_by_id = {}
        self.src_by_url = {}
        for pr in prs:
            self.pr_by_id[pr.nr] = pr
            self.src_by_url[f"{pr.source.owner}/{pr.source.slug}"] = pr.source

    def response(self):
        s = ""
        for pr in self.prs:
            s += self.template % {
                "description": pr.description,
                "owner": self.owner,
                "slug": self.slug,
                "display_name": pr.display_name,
                "title": pr.title,
                "hash": pr.source.hash,
                "short_hash": pr.source.hash[0:12],
                "src_owner": pr.source.owner,
                "src_slug": pr.source.slug,
                "created_on": pr.created_on,
                "updated_on": pr.updated_on,
                "id": pr.nr,
            }
        return f"""\
{{

    "pagelen": 10,
    "values": [{s}],
    "page": 1

}}
"""

    def getPage(self, url, timeout=None, headers=None):
        list_url_re = re.compile(
            f"https://api.bitbucket.org/2.0/repositories/{self.owner}/{self.slug}/pullrequests"
        )
        pr_url_re = re.compile(
            rf"https://api.bitbucket.org/2.0/repositories/{self.owner}/{self.slug}/pullrequests/(?P<id>\d+)"
        )
        source_commit_url_re = re.compile(
            r"https://api.bitbucket.org/2.0/repositories/(?P<src_owner>.*)/(?P<src_slug>.*)/commit/(?P<hash>\d+)"
        )
        source_url_re = re.compile(
            r"https://api.bitbucket.org/2.0/repositories/(?P<src_owner>.*)/(?P<src_slug>.*)"
        )

        if list_url_re.match(url):
            return defer.succeed(self.request())

        m = pr_url_re.match(url)
        if m:
            return self.pr_by_id[int(m.group("id"))].request()

        m = source_commit_url_re.match(url)
        if m:
            return self.src_by_url[f'{m.group("src_owner")}/{m.group("src_slug")}'].request()

        m = source_url_re.match(url)
        if m:
            return self.src_by_url[f'{m.group("src_owner")}/{m.group("src_slug")}'].repo_request()

        raise Error(code=404)


class TestBitbucketPullrequestPoller(
    changesource.ChangeSourceMixin, TestReactorMixin, LoggingMixin, unittest.TestCase
):
    def setUp(self):
        self.setup_test_reactor()
        self.setUpLogging()

        # create pull requests
        self.date = "2013-10-15T20:38:20.001797+00:00"
        self.date_epoch = datetime.strptime(
            self.date.split('.', maxsplit=1)[0], '%Y-%m-%dT%H:%M:%S'
        )
        self.rest_src = SourceRest(
            owner="contributor",
            slug="slug",
            hash="1111111111111111111111111111111111111111",
            date=self.date,
        )
        self.rest_pr = PullRequestRest(
            nr=1,
            title="title",
            description="description",
            display_name="contributor",
            source=self.rest_src,
            created_on=self.date,
        )
        self.rest_pr_list = PullRequestListRest(
            owner="owner",
            slug="slug",
            prs=[self.rest_pr],
        )

        return self.setUpChangeSource()

    def _fakeGetPage(self, result):
        # Install a fake getPage that puts the requested URL in self.getPage_got_url
        # and return result
        self.getPage_got_url = None

        def fake(url, timeout=None, headers=None):
            self.getPage_got_url = url
            return defer.succeed(result)

        self.patch(self.changesource, "getPage", fake)

    def _fakeGetPage403(self, expected_headers):
        def fail_unauthorized(url, timeout=None, headers=None):
            if headers != expected_headers:
                raise Error(code=403)

        self.patch(self.changesource, "getPage", fail_unauthorized)

    def _fakeGetPage404(self):
        def fail(url, timeout=None, headers=None):
            raise Error(code=404)

        self.patch(self.changesource, "getPage", fail)

    @defer.inlineCallbacks
    def _new_change_source(self, **kwargs):
        self._http = yield fakehttpclientservice.HTTPClientService.getService(
            self.master, self, 'https://api.bitbucket.org/2.0', auth=None
        )

        change_source = BitbucketPullrequestPoller(**kwargs)
        yield self.attachChangeSource(change_source)
        return change_source

    # tests
    @defer.inlineCallbacks
    def test_describe(self):
        yield self._new_change_source(owner='owner', slug='slug')
        assert re.search(r'owner/slug', self.changesource.describe())

    @defer.inlineCallbacks
    def test_poll_unknown_repo(self):
        # Polling a non-existent repository should result in a 404
        yield self._new_change_source(owner='owner', slug='slug')

        self._http.expect('get', '/repositories/owner/slug/pullrequests', content_json={}, code=404)

        yield self.changesource.poll()

        self.assertLogged('error 404 while loading')

    @defer.inlineCallbacks
    def test_poll_no_pull_requests(self):
        yield self._new_change_source(owner='owner', slug='slug')

        rest_pr_list = PullRequestListRest(
            owner="owner",
            slug="slug",
            prs=[],
        )

        self._http.expect(
            'get', '/repositories/owner/slug/pullrequests', content=rest_pr_list.response()
        )

        yield self.changesource.poll()

        self.assertEqual(len(self.master.data.updates.changesAdded), 0)

    @defer.inlineCallbacks
    def test_poll_new_pull_requests(self):
        yield self._new_change_source(owner='owner', slug='slug')

        self._http.expect(
            'get', '/repositories/owner/slug/pullrequests', content=self.rest_pr_list.response()
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/owner/slug/pullrequests/1',
            content=self.rest_pr.response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug/commit/111111111111',
            content=self.rest_src.response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug/diff/1111111111111111111111111111111111111111',
            content=self.rest_src.diff_response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug',
            content=self.rest_src.repo_response(),
        )

        yield self.changesource.poll()

        self.assertEqual(
            self.master.data.updates.changesAdded,
            [
                {
                    'author': 'contributor',
                    'committer': None,
                    'branch': 'default',
                    'category': None,
                    'codebase': None,
                    'comments': 'pull-request #1: title\nhttps://bitbucket.org/owner/slug/pull-request/1',
                    'files': ['path/to/a/file.txt'],
                    'project': '',
                    'properties': {
                        'pullrequesturl': 'https://bitbucket.org/owner/slug/pull-request/1'
                    },
                    'repository': 'https://bitbucket.org/contributor/slug',
                    'revision': '1111111111111111111111111111111111111111',
                    'revlink': 'https://bitbucket.org/contributor/slug/commits/111111111111',
                    'src': 'bitbucket',
                    'when_timestamp': 1381869500,
                }
            ],
        )

    @defer.inlineCallbacks
    def test_poll_no_updated_pull_request(self):
        yield self._new_change_source(owner='owner', slug='slug')

        self._http.expect(
            'get', '/repositories/owner/slug/pullrequests', content=self.rest_pr_list.response()
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/owner/slug/pullrequests/1',
            content=self.rest_pr.response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug/commit/111111111111',
            content=self.rest_src.response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug/diff/1111111111111111111111111111111111111111',
            content=self.rest_src.diff_response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug',
            content=self.rest_src.repo_response(),
        )

        self._http.expect(
            'get', '/repositories/owner/slug/pullrequests', content=self.rest_pr_list.response()
        )

        yield self.changesource.poll()

        self.assertEqual(
            self.master.data.updates.changesAdded,
            [
                {
                    'author': 'contributor',
                    'committer': None,
                    'branch': 'default',
                    'category': None,
                    'codebase': None,
                    'comments': 'pull-request #1: title\nhttps://bitbucket.org/owner/slug/pull-request/1',
                    'files': ['path/to/a/file.txt'],
                    'project': '',
                    'properties': {
                        'pullrequesturl': 'https://bitbucket.org/owner/slug/pull-request/1'
                    },
                    'repository': 'https://bitbucket.org/contributor/slug',
                    'revision': '1111111111111111111111111111111111111111',
                    'revlink': 'https://bitbucket.org/contributor/slug/commits/111111111111',
                    'src': 'bitbucket',
                    'when_timestamp': 1381869500,
                }
            ],
        )

        # repoll
        yield self.changesource.poll()
        self.assertEqual(len(self.master.data.updates.changesAdded), 1)

    @defer.inlineCallbacks
    def test_poll_updated_pull_request(self):
        yield self._new_change_source(owner='owner', slug='slug')

        rest_src2 = SourceRest(
            owner="contributor",
            slug="slug",
            hash="2222222222222222222222222222222222222222",
            date=self.date,
        )
        rest_pr2 = PullRequestRest(
            nr=1,
            title="title",
            description="description",
            display_name="contributor",
            source=rest_src2,
            created_on=self.date,
        )

        rest_pr_list2 = PullRequestListRest(
            owner="owner",
            slug="slug",
            prs=[rest_pr2],
        )

        self._http.expect(
            'get', '/repositories/owner/slug/pullrequests', content=self.rest_pr_list.response()
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/owner/slug/pullrequests/1',
            content=self.rest_pr.response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug/commit/111111111111',
            content=self.rest_src.response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug/diff/1111111111111111111111111111111111111111',
            content=self.rest_src.diff_response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug',
            content=self.rest_src.repo_response(),
        )

        self._http.expect(
            'get', '/repositories/owner/slug/pullrequests', content=rest_pr_list2.response()
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/owner/slug/pullrequests/1',
            content=rest_pr2.response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug/commit/222222222222',
            content=rest_src2.response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug/diff/2222222222222222222222222222222222222222',
            content=self.rest_src.diff_response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug',
            content=rest_src2.repo_response(),
        )

        yield self.changesource.poll()
        self.maxDiff = None
        self.assertEqual(
            self.master.data.updates.changesAdded,
            [
                {
                    'author': 'contributor',
                    'committer': None,
                    'branch': 'default',
                    'category': None,
                    'codebase': None,
                    'comments': 'pull-request #1: title\nhttps://bitbucket.org/owner/slug/pull-request/1',
                    'files': ['path/to/a/file.txt'],
                    'project': '',
                    'properties': {
                        'pullrequesturl': 'https://bitbucket.org/owner/slug/pull-request/1'
                    },
                    'repository': 'https://bitbucket.org/contributor/slug',
                    'revision': '1111111111111111111111111111111111111111',
                    'revlink': 'https://bitbucket.org/contributor/slug/commits/111111111111',
                    'src': 'bitbucket',
                    'when_timestamp': 1381869500,
                }
            ],
        )

        yield self.changesource.poll()

        self.assertEqual(
            self.master.data.updates.changesAdded,
            [
                {
                    'author': 'contributor',
                    'committer': None,
                    'branch': 'default',
                    'category': None,
                    'codebase': None,
                    'comments': 'pull-request #1: title\nhttps://bitbucket.org/owner/slug/pull-request/1',
                    'files': ['path/to/a/file.txt'],
                    'project': '',
                    'properties': {
                        'pullrequesturl': 'https://bitbucket.org/owner/slug/pull-request/1'
                    },
                    'repository': 'https://bitbucket.org/contributor/slug',
                    'revision': '1111111111111111111111111111111111111111',
                    'revlink': 'https://bitbucket.org/contributor/slug/commits/111111111111',
                    'src': 'bitbucket',
                    'when_timestamp': 1381869500,
                },
                {
                    'author': 'contributor',
                    'committer': None,
                    'branch': 'default',
                    'category': None,
                    'codebase': None,
                    'comments': 'pull-request #1: title\nhttps://bitbucket.org/owner/slug/pull-request/1',
                    'files': ['path/to/a/file.txt'],
                    'project': '',
                    'properties': {
                        'pullrequesturl': 'https://bitbucket.org/owner/slug/pull-request/1'
                    },
                    'repository': 'https://bitbucket.org/contributor/slug',
                    'revision': '2222222222222222222222222222222222222222',
                    'revlink': 'https://bitbucket.org/contributor/slug/commits/222222222222',
                    'src': 'bitbucket',
                    'when_timestamp': 1381869500,
                },
            ],
        )

    @defer.inlineCallbacks
    def test_poll_pull_request_filter_False(self):
        yield self._new_change_source(
            owner='owner', slug='slug', pullrequest_filter=lambda x: False
        )

        self._http.expect(
            'get', '/repositories/owner/slug/pullrequests', content=self.rest_pr_list.response()
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/owner/slug/pullrequests/1',
            content=self.rest_pr.response(),
        )

        yield self.changesource.poll()

        self.assertEqual(len(self.master.data.updates.changesAdded), 0)

    @defer.inlineCallbacks
    def test_poll_pull_request_filter_True(self):
        yield self._new_change_source(owner='owner', slug='slug', pullrequest_filter=lambda x: True)

        self._http.expect(
            'get', '/repositories/owner/slug/pullrequests', content=self.rest_pr_list.response()
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/owner/slug/pullrequests/1',
            content=self.rest_pr.response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug/commit/111111111111',
            content=self.rest_src.response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug/diff/1111111111111111111111111111111111111111',
            content=self.rest_src.diff_response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug',
            content=self.rest_src.repo_response(),
        )

        yield self.changesource.poll()

        self.assertEqual(
            self.master.data.updates.changesAdded,
            [
                {
                    'author': 'contributor',
                    'committer': None,
                    'branch': 'default',
                    'category': None,
                    'codebase': None,
                    'comments': 'pull-request #1: title\nhttps://bitbucket.org/owner/slug/pull-request/1',
                    'files': ['path/to/a/file.txt'],
                    'project': '',
                    'properties': {
                        'pullrequesturl': 'https://bitbucket.org/owner/slug/pull-request/1'
                    },
                    'repository': 'https://bitbucket.org/contributor/slug',
                    'revision': '1111111111111111111111111111111111111111',
                    'revlink': 'https://bitbucket.org/contributor/slug/commits/111111111111',
                    'src': 'bitbucket',
                    'when_timestamp': 1381869500,
                }
            ],
        )

    @defer.inlineCallbacks
    def test_poll_pull_request_not_useTimestamps(self):
        yield self._new_change_source(owner='owner', slug='slug', useTimestamps=False)

        self._http.expect(
            'get', '/repositories/owner/slug/pullrequests', content=self.rest_pr_list.response()
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/owner/slug/pullrequests/1',
            content=self.rest_pr.response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug/commit/111111111111',
            content=self.rest_src.response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug/diff/1111111111111111111111111111111111111111',
            content=self.rest_src.diff_response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug',
            content=self.rest_src.repo_response(),
        )

        self.reactor.advance(1396825656)

        yield self.changesource.poll()
        self.assertEqual(
            self.master.data.updates.changesAdded,
            [
                {
                    'author': 'contributor',
                    'committer': None,
                    'branch': 'default',
                    'category': None,
                    'codebase': None,
                    'comments': 'pull-request #1: title\nhttps://bitbucket.org/owner/slug/pull-request/1',
                    'files': ['path/to/a/file.txt'],
                    'project': '',
                    'properties': {
                        'pullrequesturl': 'https://bitbucket.org/owner/slug/pull-request/1'
                    },
                    'repository': 'https://bitbucket.org/contributor/slug',
                    'revision': '1111111111111111111111111111111111111111',
                    'revlink': 'https://bitbucket.org/contributor/slug/commits/111111111111',
                    'src': 'bitbucket',
                    'when_timestamp': 1396825656,
                }
            ],
        )

    @defer.inlineCallbacks
    def test_poll_pull_request_properties(self):
        yield self._new_change_source(
            owner='owner', slug='slug', bitbucket_property_whitelist=["bitbucket.*"]
        )

        self._http.expect(
            'get', '/repositories/owner/slug/pullrequests', content=self.rest_pr_list.response()
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/owner/slug/pullrequests/1',
            content=self.rest_pr.response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug/commit/111111111111',
            content=self.rest_src.response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug/diff/1111111111111111111111111111111111111111',
            content=self.rest_src.diff_response(),
        )

        self._http.expect(
            'get',
            'https://api.bitbucket.org/2.0/repositories/contributor/slug',
            content=self.rest_src.repo_response(),
        )

        yield self.changesource.poll()
        self.assertEqual(
            self.master.data.updates.changesAdded,
            [
                {
                    'author': 'contributor',
                    'committer': None,
                    'branch': 'default',
                    'category': None,
                    'codebase': None,
                    'comments': 'pull-request #1: title\nhttps://bitbucket.org/owner/slug/pull-request/1',
                    'files': ['path/to/a/file.txt'],
                    'project': '',
                    'properties': {
                        'pullrequesturl': 'https://bitbucket.org/owner/slug/pull-request/1',
                        'bitbucket.author.display_name': 'contributor',
                        'bitbucket.created_on': '2013-10-15T20:38:20.001797+00:00',
                        'bitbucket.description': 'description',
                        'bitbucket.id': 1,
                        'bitbucket.links.html.href': 'https://bitbucket.org/owner/slug/pull-request/1',
                        'bitbucket.links.self.href': 'https://api.bitbucket.org/2.0/'
                        'repositories/owner/slug/pullrequests/1',
                        'bitbucket.merge_commit': None,
                        'bitbucket.source.branch.name': 'default',
                        'bitbucket.source.commit.hash': '111111111111',
                        'bitbucket.source.commit.links.self.href': 'https://api.bitbucket.org/2.0/'
                        'repositories/contributor/slug/'
                        'commit/111111111111',
                        'bitbucket.source.repository.links.self.href': 'https://api.bitbucket.org/2.0/'
                        'repositories/contributor/slug',
                        'bitbucket.state': 'OPEN',
                        'bitbucket.title': 'title',
                        'bitbucket.updated_on': '2013-10-15T20:38:20.001797+00:00',
                    },
                    'repository': 'https://bitbucket.org/contributor/slug',
                    'revision': '1111111111111111111111111111111111111111',
                    'revlink': 'https://bitbucket.org/contributor/slug/commits/111111111111',
                    'src': 'bitbucket',
                    'when_timestamp': 1381869500,
                }
            ],
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
#
# copied from foolscap


from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import log


class _SimpleCallQueue:
    _reactor = reactor

    def __init__(self):
        self._events = []
        self._flushObservers = []
        self._timer = None
        self._in_turn = False

    def append(self, cb, args, kwargs):
        self._events.append((cb, args, kwargs))
        if not self._timer:
            self._timer = self._reactor.callLater(0, self._turn)

    def _turn(self):
        self._timer = None
        self._in_turn = True
        # flush all the messages that are currently in the queue. If anything
        # gets added to the queue while we're doing this, those events will
        # be put off until the next turn.
        events = self._events
        self._events = []
        for cb, args, kwargs in events:
            try:
                cb(*args, **kwargs)
            except Exception:
                log.err()
        self._in_turn = False
        if self._events and not self._timer:
            self._timer = self._reactor.callLater(0, self._turn)
        if not self._events:
            observers = self._flushObservers
            self._flushObservers = []
            for o in observers:
                o.callback(None)

    def flush(self):
        if not self._events and not self._in_turn:
            return defer.succeed(None)
        d = defer.Deferred()
        self._flushObservers.append(d)
        return d


_theSimpleQueue = _SimpleCallQueue()


def eventually(cb, *args, **kwargs):
    _theSimpleQueue.append(cb, args, kwargs)


def fireEventually(value=None):
    d = defer.Deferred()
    eventually(d.callback, value)
    return d


def flushEventualQueue(_ignored=None):
    return _theSimpleQueue.flush()


def _setReactor(r=None):
    # This sets the reactor used to schedule future events to r.  If r is None
    # (the default), the reactor is reset to its default value.
    # This should only be used for unit tests.
    if r is None:
        r = reactor
    _theSimpleQueue._reactor = r



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os
import unittest

from twisted.python import util


def onlyOnce(fn):
    "Set up FN to only run once within an interpreter instance"

    def wrap(*args, **kwargs):
        if hasattr(fn, 'called'):
            return None
        fn.called = 1
        return fn(*args, **kwargs)

    util.mergeFunctionMetadata(fn, wrap)
    return wrap


# NOTE: all of these patches test for applicability *before* importing the
# patch module.  This will help cut down on unnecessary imports where the
# patches are not needed, and also avoid problems with patches importing
# private things in external libraries that no longer exist.


@onlyOnce
def patch_testcase_timeout():
    # any test that should take more than 5 second should be annotated so.
    unittest.TestCase.timeout = 5

    # but we know that the DB tests are very slow, so we increase a bit that value for
    # real database tests
    if os.environ.get("BUILDBOT_TEST_DB_URL", None) is not None:
        unittest.TestCase.timeout = 120


@onlyOnce
def patch_servicechecks():
    from buildbot.monkeypatches import servicechecks

    servicechecks.patch()


@onlyOnce
def patch_decorators():
    from buildbot.monkeypatches import decorators

    decorators.patch()


@onlyOnce
def patch_config_for_unit_tests():
    from buildbot.config.master import set_is_in_unit_tests

    # by default, buildbot.config warns about not configured buildbotNetUsageData.
    # its important for users to not leak information, but unneeded and painful for tests
    set_is_in_unit_tests(True)


def patch_all(for_tests=False):
    if for_tests:
        patch_testcase_timeout()
        patch_config_for_unit_tests()
    patch_servicechecks()
    patch_decorators()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from contextlib import contextmanager
from typing import TYPE_CHECKING
from typing import Callable
from typing import ClassVar
from typing import Sequence

from twisted.python import deprecate
from twisted.python import versions

from buildbot import interfaces
from buildbot import util
from buildbot.process import buildstep
from buildbot.process.build import Build
from buildbot.steps.download_secret_to_worker import DownloadSecretsToWorker
from buildbot.steps.download_secret_to_worker import RemoveWorkerFileSecret
from buildbot.steps.shell import Compile
from buildbot.steps.shell import Configure
from buildbot.steps.shell import PerlModuleTest
from buildbot.steps.shell import ShellCommand
from buildbot.steps.shell import Test
from buildbot.steps.source.cvs import CVS
from buildbot.steps.source.svn import SVN
from buildbot.warnings import warn_deprecated

if TYPE_CHECKING:
    from buildbot.process.builder import Builder


# deprecated, use BuildFactory.addStep
@deprecate.deprecated(versions.Version("buildbot", 0, 8, 6))
def s(steptype, **kwargs):
    # convenience function for master.cfg files, to create step
    # specification tuples
    return buildstep.get_factory_from_step_or_factory(steptype(**kwargs))


class BuildFactory(util.ComparableMixin):
    """
    @cvar  buildClass: class to use when creating builds
    @type  buildClass: L{buildbot.process.build.Build}
    """

    buildClass: type[Build] = Build
    useProgress = True
    workdir = "build"
    compare_attrs: ClassVar[Sequence[str]] = ('buildClass', 'steps', 'useProgress', 'workdir')

    def __init__(self, steps=None):
        self.steps = []
        if steps:
            self.addSteps(steps)

    def newBuild(self, requests, builder: Builder) -> Build:
        """Create a new Build instance.

        @param requests: a list of buildrequest dictionaries describing what is
        to be built
        """
        b = self.buildClass(requests, builder)
        b.useProgress = self.useProgress
        b.workdir = self.workdir
        if callable(self.workdir):
            warn_deprecated(
                '4.3.0',
                'BuildFactory workdir callable support has been deprecated. Use renderables instead',
            )
        b.setStepFactories(self.steps)
        return b

    def addStep(self, step):
        if not interfaces.IBuildStep.providedBy(
            step
        ) and not interfaces.IBuildStepFactory.providedBy(step):
            raise TypeError('step must be an instance of a BuildStep')
        self.steps.append(buildstep.get_factory_from_step_or_factory(step))

    def addSteps(self, steps, withSecrets=None):
        if withSecrets is None:
            withSecrets = []
        if withSecrets:
            self.addStep(DownloadSecretsToWorker(withSecrets))
        for s in steps:
            self.addStep(s)
        if withSecrets:
            self.addStep(RemoveWorkerFileSecret(withSecrets))

    def setSkipBuildIf(self, predicate: Callable[[Build], bool]):
        self.skipBuildIf = predicate

    @contextmanager
    def withSecrets(self, secrets):
        self.addStep(DownloadSecretsToWorker(secrets))
        yield self
        self.addStep(RemoveWorkerFileSecret(secrets))


# BuildFactory subclasses for common build tools


class _DefaultCommand:
    # Used to indicate a default command to the step.
    pass


class GNUAutoconf(BuildFactory):
    def __init__(
        self,
        source,
        configure="./configure",
        configureEnv=None,
        configureFlags=None,
        reconf=None,
        compile=_DefaultCommand,
        test=_DefaultCommand,
        distcheck=_DefaultCommand,
    ):
        if configureEnv is None:
            configureEnv = {}
        if configureFlags is None:
            configureFlags = []
        if compile is _DefaultCommand:
            compile = ["make", "all"]
        if test is _DefaultCommand:
            test = ["make", "check"]
        if distcheck is _DefaultCommand:
            distcheck = ["make", "distcheck"]

        super().__init__([source])

        if reconf is True:
            reconf = ["autoreconf", "-si"]
        if reconf is not None:
            self.addStep(ShellCommand(name="autoreconf", command=reconf, env=configureEnv))

        if configure is not None:
            # we either need to wind up with a string (which will be
            # space-split), or with a list of strings (which will not). The
            # list of strings is the preferred form.
            if isinstance(configure, str):
                if configureFlags:
                    assert " " not in configure  # please use list instead
                    command = [configure, *configureFlags]
                else:
                    command = configure
            else:
                assert isinstance(configure, (list, tuple))
                command = configure + configureFlags
            self.addStep(Configure(command=command, env=configureEnv))
        if compile is not None:
            self.addStep(Compile(command=compile, env=configureEnv))
        if test is not None:
            self.addStep(Test(command=test, env=configureEnv))
        if distcheck is not None:
            self.addStep(Test(command=distcheck, env=configureEnv))


class CPAN(BuildFactory):
    def __init__(self, source, perl="perl"):
        super().__init__([source])
        self.addStep(Configure(command=[perl, "Makefile.PL"]))
        self.addStep(Compile(command=["make"]))
        self.addStep(PerlModuleTest(command=["make", "test"]))


# deprecated, use Distutils
@deprecate.deprecated(versions.Version("buildbot", 4, 0, 0))
class Distutils(BuildFactory):
    def __init__(self, source, python="python", test=None):
        super().__init__([source])
        self.addStep(Compile(command=[python, "./setup.py", "build"]))
        if test is not None:
            self.addStep(Test(command=test))


class Trial(BuildFactory):
    """Build a python module that uses distutils and trial. Set 'tests' to
    the module in which the tests can be found, or set useTestCaseNames=True
    to always have trial figure out which tests to run (based upon which
    files have been changed).

    See docs/factories.xhtml for usage samples. Not all of the Trial
    BuildStep options are available here, only the most commonly used ones.
    To get complete access, you will need to create a custom
    BuildFactory."""

    trial = "trial"
    randomly = False
    recurse = False

    def __init__(
        self,
        source,
        buildpython=None,
        trialpython=None,
        trial=None,
        testpath=".",
        randomly=None,
        recurse=None,
        tests=None,
        useTestCaseNames=False,
        env=None,
    ):
        super().__init__([source])
        assert tests or useTestCaseNames, "must use one or the other"
        if buildpython is None:
            buildpython = ["python"]
        if trialpython is None:
            trialpython = []
        if trial is not None:
            self.trial = trial
        if randomly is not None:
            self.randomly = randomly
        if recurse is not None:
            self.recurse = recurse

        from buildbot.steps.python_twisted import Trial

        buildcommand = [*buildpython, "./setup.py", "build"]
        self.addStep(Compile(command=buildcommand, env=env))
        self.addStep(
            Trial(
                python=trialpython,
                trial=self.trial,
                testpath=testpath,
                tests=tests,
                testChanges=useTestCaseNames,
                randomly=self.randomly,
                recurse=self.recurse,
                env=env,
            )
        )


# compatibility classes, will go away. Note that these only offer
# compatibility at the constructor level: if you have subclassed these
# factories, your subclasses are unlikely to still work correctly.

ConfigurableBuildFactory = BuildFactory


class BasicBuildFactory(GNUAutoconf):
    # really a "GNU Autoconf-created tarball -in-CVS tree" builder

    def __init__(
        self,
        cvsroot,
        cvsmodule,
        configure=None,
        configureEnv=None,
        compile="make all",
        test="make check",
        cvsCopy=False,
    ):
        if configureEnv is None:
            configureEnv = {}
        mode = "full"
        method = "clobber"
        if cvsCopy:
            method = "copy"
        source = CVS(cvsroot=cvsroot, cvsmodule=cvsmodule, mode=mode, method=method)
        super().__init__(
            source, configure=configure, configureEnv=configureEnv, compile=compile, test=test
        )


class QuickBuildFactory(BasicBuildFactory):
    useProgress = False

    def __init__(
        self,
        cvsroot,
        cvsmodule,
        configure=None,
        configureEnv=None,
        compile="make all",
        test="make check",
        cvsCopy=False,
    ):
        if configureEnv is None:
            configureEnv = {}
        mode = "incremental"
        source = CVS(cvsroot=cvsroot, cvsmodule=cvsmodule, mode=mode)
        super().__init__(
            source, configure=configure, configureEnv=configureEnv, compile=compile, test=test
        )


class BasicSVN(GNUAutoconf):
    def __init__(
        self, svnurl, configure=None, configureEnv=None, compile="make all", test="make check"
    ):
        if configureEnv is None:
            configureEnv = {}
        source = SVN(svnurl=svnurl, mode="incremental")
        super().__init__(
            source, configure=configure, configureEnv=configureEnv, compile=compile, test=test
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import datetime

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.db import buildrequests
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import db
from buildbot.util import UTC
from buildbot.util import epoch2datetime


class Tests(TestReactorMixin, unittest.TestCase):
    # test that the datetime translations are done correctly by specifying
    # the epoch timestamp and datetime objects explicitly.  These should
    # pass regardless of the local timezone used while running tests!
    CLAIMED_AT = datetime.datetime(1978, 6, 15, 12, 31, 15, tzinfo=UTC)
    CLAIMED_AT_EPOCH = 266761875
    SUBMITTED_AT = datetime.datetime(1979, 6, 15, 12, 31, 15, tzinfo=UTC)
    SUBMITTED_AT_EPOCH = 298297875
    COMPLETE_AT = datetime.datetime(1980, 6, 15, 12, 31, 15, tzinfo=UTC)
    COMPLETE_AT_EPOCH = 329920275
    BSID = 567
    BLDRID1 = 890
    BLDRID2 = 891
    BLDRID3 = 893
    MASTER_ID = "set in setUp"
    OTHER_MASTER_ID = "set in setUp"

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True)

        # set up a sourcestamp and buildset for use below
        self.MASTER_ID = fakedb.FakeDBConnector.MASTER_ID
        self.OTHER_MASTER_ID = self.MASTER_ID + 1111
        self.master.masterid = self.MASTER_ID
        self.db = self.master.db

        yield self.master.db.insert_test_data([
            fakedb.SourceStamp(id=234),
            fakedb.Master(id=self.MASTER_ID, name="fake master"),
            fakedb.Master(id=self.OTHER_MASTER_ID, name="other"),
            fakedb.Buildset(id=self.BSID),
            fakedb.Builder(id=self.BLDRID1, name="builder1"),
            fakedb.Builder(id=self.BLDRID2, name="builder2"),
            fakedb.Builder(id=self.BLDRID3, name="builder3"),
            fakedb.BuildsetSourceStamp(buildsetid=self.BSID, sourcestampid=234),
        ])

    @defer.inlineCallbacks
    def test_getBuildRequest(self):
        yield self.master.db.insert_test_data([
            fakedb.BuildRequest(
                id=44,
                buildsetid=self.BSID,
                builderid=self.BLDRID1,
                complete=1,
                results=75,
                priority=7,
                submitted_at=self.SUBMITTED_AT_EPOCH,
                complete_at=self.COMPLETE_AT_EPOCH,
            ),
            fakedb.BuildRequestClaim(
                brid=44, masterid=self.MASTER_ID, claimed_at=self.CLAIMED_AT_EPOCH
            ),
        ])
        brdict = yield self.db.buildrequests.getBuildRequest(44)

        yield self.assertEqual(
            brdict,
            buildrequests.BuildRequestModel(
                buildrequestid=44,
                buildsetid=self.BSID,
                builderid=self.BLDRID1,
                buildername="builder1",
                priority=7,
                claimed_by_masterid=self.MASTER_ID,
                complete=True,
                results=75,
                claimed_at=self.CLAIMED_AT,
                submitted_at=self.SUBMITTED_AT,
                complete_at=self.COMPLETE_AT,
                waited_for=False,
            ),
        )

    @defer.inlineCallbacks
    def test_getBuildRequest_missing(self):
        brdict = yield self.db.buildrequests.getBuildRequest(44)

        self.assertEqual(brdict, None)

    @defer.inlineCallbacks
    def do_test_getBuildRequests_claim_args(self, **kwargs):
        expected = kwargs.pop('expected')
        yield self.master.db.insert_test_data([
            # 50: claimed by this master
            fakedb.BuildRequest(id=50, buildsetid=self.BSID, builderid=self.BLDRID1),
            fakedb.BuildRequestClaim(
                brid=50, masterid=self.MASTER_ID, claimed_at=self.CLAIMED_AT_EPOCH
            ),
            # 51: claimed by another master
            fakedb.BuildRequest(id=51, buildsetid=self.BSID, builderid=self.BLDRID2),
            fakedb.BuildRequestClaim(
                brid=51, masterid=self.OTHER_MASTER_ID, claimed_at=self.CLAIMED_AT_EPOCH
            ),
            # 52: unclaimed
            fakedb.BuildRequest(id=52, buildsetid=self.BSID, builderid=self.BLDRID1),
            # 53: unclaimed but complete (should not appear for claimed=False)
            fakedb.BuildRequest(id=53, buildsetid=self.BSID, builderid=self.BLDRID1, complete=1),
        ])
        brlist = yield self.db.buildrequests.getBuildRequests(**kwargs)

        self.assertEqual(sorted([br.buildrequestid for br in brlist]), sorted(expected))

    def test_getBuildRequests_no_claimed_arg(self):
        return self.do_test_getBuildRequests_claim_args(expected=[50, 51, 52, 53])

    def test_getBuildRequests_claimed_mine(self):
        return self.do_test_getBuildRequests_claim_args(claimed=self.MASTER_ID, expected=[50])

    def test_getBuildRequests_claimed_true(self):
        return self.do_test_getBuildRequests_claim_args(claimed=True, expected=[50, 51])

    def test_getBuildRequests_unclaimed(self):
        return self.do_test_getBuildRequests_claim_args(claimed=False, expected=[52])

    @defer.inlineCallbacks
    def do_test_getBuildRequests_buildername_arg(self, **kwargs):
        expected = kwargs.pop('expected')
        yield self.master.db.insert_test_data([
            # 8: 'bb'
            fakedb.BuildRequest(id=8, buildsetid=self.BSID, builderid=self.BLDRID1),
            # 9: 'cc'
            fakedb.BuildRequest(id=9, buildsetid=self.BSID, builderid=self.BLDRID2),
            # 10: 'cc'
            fakedb.BuildRequest(id=10, buildsetid=self.BSID, builderid=self.BLDRID2),
        ])
        brlist = yield self.db.buildrequests.getBuildRequests(**kwargs)

        self.assertEqual(sorted([br.buildrequestid for br in brlist]), sorted(expected))

    @defer.inlineCallbacks
    def do_test_getBuildRequests_complete_arg(self, **kwargs):
        expected = kwargs.pop('expected')
        yield self.master.db.insert_test_data([
            # 70: incomplete
            fakedb.BuildRequest(
                id=70, buildsetid=self.BSID, builderid=self.BLDRID1, complete=0, complete_at=None
            ),
            # 80: complete
            fakedb.BuildRequest(
                id=80,
                buildsetid=self.BSID,
                builderid=self.BLDRID1,
                complete=1,
                complete_at=self.COMPLETE_AT_EPOCH,
            ),
            # 81: complete but no complete_at
            fakedb.BuildRequest(
                id=81, buildsetid=self.BSID, builderid=self.BLDRID1, complete=1, complete_at=0
            ),
            # 82: complete_at set but complete is false, so not complete
            fakedb.BuildRequest(
                id=82,
                buildsetid=self.BSID,
                builderid=self.BLDRID1,
                complete=0,
                complete_at=self.COMPLETE_AT_EPOCH,
            ),
        ])
        brlist = yield self.db.buildrequests.getBuildRequests(**kwargs)

        self.assertEqual(sorted([br.buildrequestid for br in brlist]), sorted(expected))

    def test_getBuildRequests_complete_none(self):
        return self.do_test_getBuildRequests_complete_arg(expected=[70, 80, 81, 82])

    def test_getBuildRequests_complete_true(self):
        return self.do_test_getBuildRequests_complete_arg(complete=True, expected=[80, 81])

    def test_getBuildRequests_complete_false(self):
        return self.do_test_getBuildRequests_complete_arg(complete=False, expected=[70, 82])

    @defer.inlineCallbacks
    def test_getBuildRequests_bsid_arg(self):
        yield self.master.db.insert_test_data([
            # the buildset that we are *not* looking for
            fakedb.Buildset(id=self.BSID + 1),
            fakedb.BuildRequest(
                id=70, buildsetid=self.BSID, builderid=self.BLDRID1, complete=0, complete_at=None
            ),
            fakedb.BuildRequest(
                id=71,
                buildsetid=self.BSID + 1,
                builderid=self.BLDRID1,
                complete=0,
                complete_at=None,
            ),
            fakedb.BuildRequest(
                id=72, buildsetid=self.BSID, builderid=self.BLDRID1, complete=0, complete_at=None
            ),
        ])
        brlist = yield self.db.buildrequests.getBuildRequests(bsid=self.BSID)

        self.assertEqual(sorted([br.buildrequestid for br in brlist]), sorted([70, 72]))

    @defer.inlineCallbacks
    def test_getBuildRequests_combo(self):
        yield self.master.db.insert_test_data([
            # 44: everything we want
            fakedb.BuildRequest(
                id=44,
                buildsetid=self.BSID,
                builderid=self.BLDRID1,
                complete=1,
                results=92,
                complete_at=self.COMPLETE_AT_EPOCH,
            ),
            fakedb.BuildRequestClaim(
                brid=44, masterid=self.MASTER_ID, claimed_at=self.CLAIMED_AT_EPOCH
            ),
            # 45: different builderid
            fakedb.BuildRequest(
                id=45,
                buildsetid=self.BSID,
                builderid=self.BLDRID2,
                complete=1,
                complete_at=self.COMPLETE_AT_EPOCH,
            ),
            fakedb.BuildRequestClaim(
                brid=45, masterid=self.MASTER_ID, claimed_at=self.CLAIMED_AT_EPOCH
            ),
            # 46: incomplete
            fakedb.BuildRequest(
                id=46,
                buildsetid=self.BSID,
                builderid=self.BLDRID1,
                complete=0,
                results=92,
                complete_at=0,
            ),
            fakedb.BuildRequestClaim(
                brid=46, masterid=self.MASTER_ID, claimed_at=self.CLAIMED_AT_EPOCH
            ),
            # 47: unclaimed
            fakedb.BuildRequest(
                id=47,
                buildsetid=self.BSID,
                builderid=self.BLDRID1,
                complete=1,
                results=92,
                complete_at=self.COMPLETE_AT_EPOCH,
            ),
            # 48: claimed by other
            fakedb.BuildRequest(
                id=48,
                buildsetid=self.BSID,
                builderid=self.BLDRID1,
                complete=1,
                results=92,
                complete_at=self.COMPLETE_AT_EPOCH,
            ),
            fakedb.BuildRequestClaim(
                brid=48, masterid=self.OTHER_MASTER_ID, claimed_at=self.CLAIMED_AT_EPOCH
            ),
            # 49: different bsid
            fakedb.Buildset(id=self.BSID + 1),
            fakedb.BuildRequest(
                id=49,
                buildsetid=self.BSID + 1,
                builderid=self.BLDRID1,
                complete=1,
                results=92,
                complete_at=self.COMPLETE_AT_EPOCH,
            ),
            fakedb.BuildRequestClaim(
                brid=49, masterid=self.MASTER_ID, claimed_at=self.CLAIMED_AT_EPOCH
            ),
        ])
        brlist = yield self.db.buildrequests.getBuildRequests(
            builderid=self.BLDRID1, claimed=self.MASTER_ID, complete=True, bsid=self.BSID
        )

        self.assertEqual([br.buildrequestid for br in brlist], [44])

    @defer.inlineCallbacks
    def do_test_getBuildRequests_branch_arg(self, **kwargs):
        expected = kwargs.pop('expected')
        yield self.master.db.insert_test_data([
            fakedb.Buildset(id=self.BSID + 1),
            fakedb.BuildRequest(id=70, buildsetid=self.BSID + 1, builderid=self.BLDRID1),
            fakedb.SourceStamp(id=self.BSID + 1, branch='branch_A'),
            fakedb.BuildsetSourceStamp(buildsetid=self.BSID + 1, sourcestampid=self.BSID + 1),
            fakedb.Buildset(id=self.BSID + 2),
            fakedb.BuildRequest(id=80, buildsetid=self.BSID + 2, builderid=self.BLDRID1),
            fakedb.SourceStamp(id=self.BSID + 2, repository='repository_A'),
            fakedb.BuildsetSourceStamp(buildsetid=self.BSID + 2, sourcestampid=self.BSID + 2),
            fakedb.Buildset(id=self.BSID + 3),
            fakedb.BuildRequest(id=90, buildsetid=self.BSID + 3, builderid=self.BLDRID1),
            fakedb.SourceStamp(id=self.BSID + 3, branch='branch_A', repository='repository_A'),
            fakedb.BuildsetSourceStamp(buildsetid=self.BSID + 3, sourcestampid=self.BSID + 3),
            # multiple sourcestamps on the same buildset are possible
            fakedb.SourceStamp(id=self.BSID + 4, branch='branch_B', repository='repository_B'),
            fakedb.BuildsetSourceStamp(buildsetid=self.BSID + 3, sourcestampid=self.BSID + 4),
        ])
        brlist = yield self.db.buildrequests.getBuildRequests(**kwargs)

        self.assertEqual(sorted([br.buildrequestid for br in brlist]), sorted(expected))

    def test_getBuildRequests_branch(self):
        return self.do_test_getBuildRequests_branch_arg(branch='branch_A', expected=[70, 90])

    def test_getBuildRequests_branch_empty(self):
        return self.do_test_getBuildRequests_branch_arg(branch='absent_branch', expected=[])

    def test_getBuildRequests_repository(self):
        return self.do_test_getBuildRequests_branch_arg(
            repository='repository_A', expected=[80, 90]
        )

    def test_getBuildRequests_repository_empty(self):
        return self.do_test_getBuildRequests_branch_arg(repository='absent_repository', expected=[])

    def test_getBuildRequests_repository_and_branch(self):
        return self.do_test_getBuildRequests_branch_arg(
            repository='repository_A', branch='branch_A', expected=[90]
        )

    def test_getBuildRequests_no_repository_nor_branch(self):
        return self.do_test_getBuildRequests_branch_arg(expected=[70, 80, 90])

    def failWithExpFailure(self, exc, expfailure=None):
        if not expfailure:
            raise exc
        self.flushLoggedErrors(expfailure)
        if isinstance(exc, expfailure):
            return
        raise exc

    @defer.inlineCallbacks
    def do_test_claimBuildRequests(
        self, rows, now, brids, expected=None, expfailure=None, claimed_at=None
    ):
        self.reactor.advance(now)

        try:
            yield self.master.db.insert_test_data(rows)
            yield self.db.buildrequests.claimBuildRequests(brids=brids, claimed_at=claimed_at)
            results = yield self.db.buildrequests.getBuildRequests()

            self.assertNotEqual(expected, None, "unexpected success from claimBuildRequests")
            self.assertEqual(
                sorted([(r.buildrequestid, r.claimed_at, r.claimed_by_masterid) for r in results]),
                sorted(expected),
            )
        except Exception as e:
            self.failWithExpFailure(e, expfailure)

    def test_claimBuildRequests_single(self):
        return self.do_test_claimBuildRequests(
            [
                fakedb.BuildRequest(id=44, buildsetid=self.BSID, builderid=self.BLDRID1),
            ],
            1300305712,
            [44],
            [(44, epoch2datetime(1300305712), self.MASTER_ID)],
        )

    def test_claimBuildRequests_single_explicit_claimed_at(self):
        return self.do_test_claimBuildRequests(
            [
                fakedb.BuildRequest(id=44, buildsetid=self.BSID, builderid=self.BLDRID1),
            ],
            1300305712,
            [44],
            [(44, epoch2datetime(14000000), self.MASTER_ID)],
            claimed_at=epoch2datetime(14000000),
        )

    def test_claimBuildRequests_multiple(self):
        return self.do_test_claimBuildRequests(
            [
                fakedb.BuildRequest(id=44, buildsetid=self.BSID, builderid=self.BLDRID1),
                fakedb.BuildRequest(id=45, buildsetid=self.BSID, builderid=self.BLDRID1),
                fakedb.BuildRequest(id=46, buildsetid=self.BSID, builderid=self.BLDRID1),
            ],
            1300305712,
            [44, 46],
            [
                (44, epoch2datetime(1300305712), self.MASTER_ID),
                (45, None, None),
                (46, epoch2datetime(1300305712), self.MASTER_ID),
            ],
        )

    def test_claimBuildRequests_stress(self):
        return self.do_test_claimBuildRequests(
            [
                fakedb.BuildRequest(id=id, buildsetid=self.BSID, builderid=self.BLDRID1)
                for id in range(1, 1000)
            ],
            1300305713,
            list(range(1, 1000)),
            [(id, epoch2datetime(1300305713), self.MASTER_ID) for id in range(1, 1000)],
        )

    def test_claimBuildRequests_other_master_claim(self):
        return self.do_test_claimBuildRequests(
            [
                fakedb.BuildRequest(id=44, buildsetid=self.BSID, builderid=self.BLDRID1),
                fakedb.BuildRequestClaim(
                    brid=44, masterid=self.OTHER_MASTER_ID, claimed_at=1300103810
                ),
            ],
            1300305712,
            [44],
            expfailure=buildrequests.AlreadyClaimedError,
        )

    @db.skip_for_dialect('mysql')
    @defer.inlineCallbacks
    def test_claimBuildRequests_other_master_claim_stress(self):
        yield self.do_test_claimBuildRequests(
            [
                fakedb.BuildRequest(id=id, buildsetid=self.BSID, builderid=self.BLDRID1)
                for id in range(1, 1000)
            ]
            + [
                fakedb.BuildRequest(id=1000, buildsetid=self.BSID, builderid=self.BLDRID1),
                # the fly in the ointment..
                fakedb.BuildRequestClaim(
                    brid=1000, masterid=self.OTHER_MASTER_ID, claimed_at=1300103810
                ),
            ],
            1300305712,
            list(range(1, 1001)),
            expfailure=buildrequests.AlreadyClaimedError,
        )
        results = yield self.db.buildrequests.getBuildRequests(claimed=True)

        # check that [1,1000) were not claimed, and 1000 is still claimed
        self.assertEqual(
            [(r.buildrequestid, r.claimed_by_masterid, r.claimed_at) for r in results][:10],
            [(1000, self.OTHER_MASTER_ID, epoch2datetime(1300103810))],
        )

    @defer.inlineCallbacks
    def test_claimBuildRequests_sequential(self):
        now = 120350934
        self.reactor.advance(now)

        yield self.master.db.insert_test_data([
            fakedb.BuildRequest(id=44, buildsetid=self.BSID, builderid=self.BLDRID1),
            fakedb.BuildRequest(id=45, buildsetid=self.BSID, builderid=self.BLDRID1),
        ])
        yield self.db.buildrequests.claimBuildRequests(brids=[44])
        yield self.db.buildrequests.claimBuildRequests(brids=[45])
        results = yield self.db.buildrequests.getBuildRequests(claimed=False)

        self.assertEqual(results, [])

    @defer.inlineCallbacks
    def do_test_completeBuildRequests(
        self, rows, now, expected=None, expfailure=None, brids=None, complete_at=None
    ):
        if brids is None:
            brids = [44]
        self.reactor.advance(now)

        try:
            yield self.master.db.insert_test_data(rows)
            yield self.db.buildrequests.completeBuildRequests(
                brids=brids, results=7, complete_at=complete_at
            )
            results = yield self.db.buildrequests.getBuildRequests()

            self.assertNotEqual(expected, None, "unexpected success from completeBuildRequests")
            self.assertEqual(
                sorted((r.buildrequestid, r.complete, r.results, r.complete_at) for r in results),
                sorted(expected),
            )

        except Exception as e:
            self.failWithExpFailure(e, expfailure)

    def test_completeBuildRequests(self):
        return self.do_test_completeBuildRequests(
            [
                fakedb.BuildRequest(id=44, buildsetid=self.BSID, builderid=self.BLDRID1),
                fakedb.BuildRequestClaim(brid=44, masterid=self.MASTER_ID, claimed_at=1300103810),
            ],
            1300305712,
            [(44, True, 7, epoch2datetime(1300305712))],
        )

    def test_completeBuildRequests_explicit_time(self):
        return self.do_test_completeBuildRequests(
            [
                fakedb.BuildRequest(id=44, buildsetid=self.BSID, builderid=self.BLDRID1),
                fakedb.BuildRequestClaim(brid=44, masterid=self.MASTER_ID, claimed_at=1300103810),
            ],
            1300305712,
            [(44, True, 7, epoch2datetime(999999))],
            complete_at=epoch2datetime(999999),
        )

    def test_completeBuildRequests_multiple(self):
        return self.do_test_completeBuildRequests(
            [
                fakedb.BuildRequest(id=44, buildsetid=self.BSID, builderid=self.BLDRID1),
                fakedb.BuildRequestClaim(brid=44, masterid=self.MASTER_ID, claimed_at=1300103810),
                fakedb.BuildRequest(id=45, buildsetid=self.BSID, builderid=self.BLDRID1),
                fakedb.BuildRequestClaim(
                    brid=45, masterid=self.OTHER_MASTER_ID, claimed_at=1300103811
                ),
                fakedb.BuildRequest(id=46, buildsetid=self.BSID, builderid=self.BLDRID1),
                fakedb.BuildRequestClaim(brid=46, masterid=self.MASTER_ID, claimed_at=1300103812),
            ],
            1300305712,
            [
                (44, True, 7, epoch2datetime(1300305712)),
                (45, False, -1, None),
                (46, True, 7, epoch2datetime(1300305712)),
            ],
            brids=[44, 46],
        )

    def test_completeBuildRequests_stress(self):
        return self.do_test_completeBuildRequests(
            [
                fakedb.BuildRequest(id=id, buildsetid=self.BSID, builderid=self.BLDRID1)
                for id in range(1, 280)
            ]
            + [
                fakedb.BuildRequestClaim(brid=id, masterid=self.MASTER_ID, claimed_at=1300103810)
                for id in range(1, 280)
            ],
            1300305712,
            [(id, True, 7, epoch2datetime(1300305712)) for id in range(1, 280)],
            brids=list(range(1, 280)),
        )

    def test_completeBuildRequests_multiple_notmine(self):
        # note that the requests are completed even though they are not mine!
        return self.do_test_completeBuildRequests(
            [
                # two unclaimed requests
                fakedb.BuildRequest(id=44, buildsetid=self.BSID, builderid=self.BLDRID1),
                fakedb.BuildRequest(id=45, buildsetid=self.BSID, builderid=self.BLDRID1),
                # and one claimed by another master
                fakedb.BuildRequest(id=46, buildsetid=self.BSID, builderid=self.BLDRID1),
                fakedb.BuildRequestClaim(
                    brid=46, masterid=self.OTHER_MASTER_ID, claimed_at=1300103812
                ),
            ],
            1300305712,
            [
                (44, True, 7, epoch2datetime(1300305712)),
                (45, True, 7, epoch2datetime(1300305712)),
                (46, True, 7, epoch2datetime(1300305712)),
            ],
            brids=[44, 45, 46],
        )

    def test_completeBuildRequests_already_completed(self):
        return self.do_test_completeBuildRequests(
            [
                fakedb.BuildRequest(
                    id=44,
                    buildsetid=self.BSID,
                    builderid=self.BLDRID1,
                    complete=1,
                    complete_at=1300104190,
                ),
            ],
            1300305712,
            expfailure=buildrequests.NotClaimedError,
        )

    def test_completeBuildRequests_no_such(self):
        return self.do_test_completeBuildRequests(
            [
                fakedb.BuildRequest(id=45, buildsetid=self.BSID, builderid=self.BLDRID1),
            ],
            1300305712,
            expfailure=buildrequests.NotClaimedError,
        )

    @defer.inlineCallbacks
    def do_test_unclaimMethod(self, method, expected):
        yield self.master.db.insert_test_data([
            # 44: a complete build (should not be unclaimed)
            fakedb.BuildRequest(
                id=44,
                buildsetid=self.BSID,
                builderid=self.BLDRID1,
                complete=1,
                results=92,
                complete_at=self.COMPLETE_AT_EPOCH,
            ),
            fakedb.BuildRequestClaim(
                brid=44, masterid=self.MASTER_ID, claimed_at=self.CLAIMED_AT_EPOCH
            ),
            # 45: incomplete build belonging to this incarnation
            fakedb.BuildRequest(
                id=45, buildsetid=self.BSID, builderid=self.BLDRID1, complete=0, complete_at=0
            ),
            fakedb.BuildRequestClaim(
                brid=45, masterid=self.MASTER_ID, claimed_at=self.CLAIMED_AT_EPOCH
            ),
            # 46: incomplete build belonging to another master
            fakedb.BuildRequest(
                id=46, buildsetid=self.BSID, builderid=self.BLDRID1, complete=0, complete_at=0
            ),
            fakedb.BuildRequestClaim(
                brid=46, masterid=self.OTHER_MASTER_ID, claimed_at=self.CLAIMED_AT_EPOCH
            ),
            # 47: unclaimed
            fakedb.BuildRequest(
                id=47, buildsetid=self.BSID, builderid=self.BLDRID1, complete=0, complete_at=0
            ),
            # 48: claimed by this master, but recently
            fakedb.BuildRequest(
                id=48, buildsetid=self.BSID, builderid=self.BLDRID1, complete=0, complete_at=0
            ),
            fakedb.BuildRequestClaim(
                brid=48, masterid=self.MASTER_ID, claimed_at=self.CLAIMED_AT_EPOCH - 50
            ),
            # 49: incomplete old build belonging to another master
            fakedb.BuildRequest(
                id=49, buildsetid=self.BSID, builderid=self.BLDRID1, complete=0, complete_at=0
            ),
            fakedb.BuildRequestClaim(
                brid=49, masterid=self.OTHER_MASTER_ID, claimed_at=self.CLAIMED_AT_EPOCH - 1000
            ),
        ])
        yield method()
        # just select the unclaimed requests
        results = yield self.db.buildrequests.getBuildRequests(claimed=False)

        self.assertEqual(sorted([r.buildrequestid for r in results]), sorted(expected))

    def test_unclaimBuildRequests(self):
        to_unclaim = [
            44,  # completed -> should not be unclaimed
            45,  # incomplete -> unclaimed
            46,  # from another master -> not unclaimed
            47,  # unclaimed -> still unclaimed
            48,  # claimed -> unclaimed
            49,  # another master -> not unclaimed
            50,  # no such buildrequest -> no error
        ]
        return self.do_test_unclaimMethod(
            lambda: self.db.buildrequests.unclaimBuildRequests(to_unclaim), [45, 47, 48]
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import log
from zope.interface import implementer

from buildbot_worker import util
from buildbot_worker.exceptions import AbandonChain
from buildbot_worker.interfaces import IWorkerCommand

# The following identifier should be updated each time this file is changed
command_version = "3.3"

# version history:
#  >=1.17: commands are interruptable
#  >=1.28: Arch understands 'revision', added Bazaar
#  >=1.33: Source classes understand 'retry'
#  >=1.39: Source classes correctly handle changes in branch (except Git)
#          Darcs accepts 'revision' (now all do but Git) (well, and P4Sync)
#          Arch/Baz should accept 'build-config'
#  >=1.51: (release 0.7.3)
#  >= 2.1: SlaveShellCommand now accepts 'initial_stdin', 'keep_stdin_open',
#          and 'logfiles'. It now sends 'log' messages in addition to
#          stdout/stdin/header/rc. It acquired writeStdin/closeStdin methods,
#          but these are not remotely callable yet.
#          (not externally visible: ShellCommandPP has writeStdin/closeStdin.
#          ShellCommand accepts new arguments (logfiles=, initialStdin=,
#          keepStdinOpen=) and no longer accepts stdin=)
#          (release 0.7.4)
#  >= 2.2: added monotone, uploadFile, and downloadFile (release 0.7.5)
#  >= 2.3: added bzr (release 0.7.6)
#  >= 2.4: Git understands 'revision' and branches
#  >= 2.5: workaround added for remote 'hg clone --rev REV' when hg<0.9.2
#  >= 2.6: added uploadDirectory
#  >= 2.7: added usePTY option to SlaveShellCommand
#  >= 2.8: added username and password args to SVN class
#  >= 2.9: add depth arg to SVN class
#  >= 2.10: CVS can handle 'extra_options' and 'export_options'
#  >= 2.11: Arch, Bazaar, and Monotone removed
#  >= 2.12: SlaveShellCommand no longer accepts 'keep_stdin_open'
#  >= 2.13: SlaveFileUploadCommand supports option 'keepstamp'
#  >= 2.14: RemoveDirectory can delete multiple directories
#  >= 2.15: 'interruptSignal' option is added to SlaveShellCommand
#  >= 2.16: 'sigtermTime' option is added to SlaveShellCommand
#  >= 2.16: runprocess supports obfuscation via tuples (#1748)
#  >= 2.16: listdir command added to read a directory
#  >= 3.0: new buildbot-worker package:
#    * worker-side usePTY configuration (usePTY='slave-config') support
#      dropped,
#    * remote method getSlaveInfo() renamed to getWorkerInfo().
#    * "slavesrc" command argument renamed to "workersrc" in uploadFile and
#      uploadDirectory commands.
#    * "slavedest" command argument renamed to "workerdest" in downloadFile
#      command.
#  >= 3.1: rmfile command added to remove a file
#  >= 3.2: shell command now reports failure reason in case the command timed out.
#  >= 3.3: shell command now supports max_lines parameter.


@implementer(IWorkerCommand)
class Command:
    """This class defines one command that can be invoked by the build master.
    The command is executed on the worker side, and always sends back a
    completion message when it finishes. It may also send intermediate status
    as it runs (by calling builder.sendStatus). Some commands can be
    interrupted (either by the build master or a local timeout), in which
    case the step is expected to complete normally with a status message that
    indicates an error occurred.

    These commands are used by BuildSteps on the master side. Each kind of
    BuildStep uses a single Command. The worker must implement all the
    Commands required by the set of BuildSteps used for any given build:
    this is checked at startup time.

    All Commands are constructed with the same signature:
     c = CommandClass(builder, stepid, args)
    where 'builder' is the parent WorkerForBuilder object, and 'args' is a
    dict that is interpreted per-command.

    The setup(args) method is available for setup, and is run from __init__.
    Mandatory args can be declared by listing them in the requiredArgs property.
    They will be checked before calling the setup(args) method.

    The Command is started with start(). This method must be implemented in a
    subclass, and it should return a Deferred. When your step is done, you
    should fire the Deferred (the results are not used). If the command is
    interrupted, it should fire the Deferred anyway.

    While the command runs. it may send status messages back to the
    buildmaster by calling self.sendStatus(statusdict). The statusdict is
    interpreted by the master-side BuildStep however it likes.

    A separate completion message is sent when the deferred fires, which
    indicates that the Command has finished, but does not carry any status
    data. If the Command needs to return an exit code of some sort, that
    should be sent as a regular status message before the deferred is fired .
    Once builder.commandComplete has been run, no more status messages may be
    sent.

    If interrupt() is called, the Command should attempt to shut down as
    quickly as possible. Child processes should be killed, new ones should
    not be started. The Command should send some kind of error status update,
    then complete as usual by firing the Deferred.

    .interrupted should be set by interrupt(), and can be tested to avoid
    sending multiple error status messages.

    If .running is False, the bot is shutting down (or has otherwise lost the
    connection to the master), and should not send any status messages. This
    is checked in Command.sendStatus .

    """

    # builder methods:
    #  sendStatus(list of tuples) (zero or more)
    #  commandComplete() or commandInterrupted() (one, at end)

    requiredArgs: list[str] = []
    debug = False
    interrupted = False
    # set by Builder, cleared on shutdown or when the Deferred fires
    running = False

    _reactor = reactor

    def __init__(self, protocol_command, command_id, args):
        self.protocol_command = protocol_command
        self.command_id = command_id  # just for logging
        self.args = args
        self.startTime = None

        missingArgs = [arg for arg in self.requiredArgs if arg not in args]
        if missingArgs:
            raise ValueError(
                "{} is missing args: {}".format(self.__class__.__name__, ", ".join(missingArgs))
            )
        self.setup(args)

    def log_msg(self, msg, *args):
        log.msg(f"(command {self.command_id}): {msg}", *args)

    def setup(self, args):
        """Override this in a subclass to extract items from the args dict."""

    def doStart(self):
        self.running = True
        self.startTime = util.now(self._reactor)
        d = defer.maybeDeferred(self.start)

        def commandComplete(res):
            self.sendStatus([("elapsed", util.now(self._reactor) - self.startTime)])
            self.running = False
            return res

        d.addBoth(commandComplete)
        return d

    def start(self):
        """Start the command. This method should return a Deferred that will
        fire when the command has completed. The Deferred's argument will be
        ignored.

        This method should be overridden by subclasses."""
        raise NotImplementedError("You must implement this in a subclass")

    def sendStatus(self, status):
        """Send a status update to the master."""
        if self.debug:
            self.log_msg(f"sendStatus: {status}")
        if not self.running:
            self.log_msg("would sendStatus but not .running")
            return
        self.protocol_command.send_update(status)

    def doInterrupt(self):
        self.running = False
        self.interrupt()

    def interrupt(self):
        """Override this in a subclass to allow commands to be interrupted.
        May be called multiple times, test and set self.interrupted=True if
        this matters."""

    # utility methods, mostly used by WorkerShellCommand and the like

    def _abandonOnFailure(self, rc):
        if not isinstance(rc, int):
            self.log_msg(f"weird, _abandonOnFailure was given rc={rc} ({type(rc)})")
        assert isinstance(rc, int)
        if rc != 0:
            raise AbandonChain(rc)
        return rc

    def _sendRC(self, res):
        self.sendStatus([('rc', 0)])

    def _checkAbandoned(self, why):
        self.log_msg("_checkAbandoned", why)
        why.trap(AbandonChain)
        self.log_msg(" abandoning chain", why.value)
        self.sendStatus([('rc', why.value.args[0])])
        return None



# Copyright Buildbot Team Members
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from twisted.internet import defer
from twisted.python.failure import Failure
from twisted.trial import unittest

from buildbot.util import Notifier


class TestException(Exception):
    """
    An exception thrown in tests.
    """


class Tests(unittest.TestCase):
    def test_wait(self):
        """
        Calling `Notifier.wait` returns a deferred that hasn't fired.
        """
        n = Notifier()
        self.assertNoResult(n.wait())

    def test_notify_no_waiters(self):
        """
        Calling `Notifier.notify` when there are no waiters does not
        raise.
        """
        n = Notifier()
        n.notify(object())
        # Does not raise.

    @defer.inlineCallbacks
    def test_notify_multiple_waiters(self):
        """
        If there all multiple waiters, `Notifier.notify` fires all
        the deferreds with the same value.
        """
        value = object()
        n = Notifier()
        ds = [n.wait(), n.wait()]
        n.notify(value)

        self.assertEqual((yield ds[0]), value)
        self.assertEqual((yield ds[1]), value)

    @defer.inlineCallbacks
    def test_new_waiters_not_notified(self):
        """
        If a new waiter is added while notifying, it won't be
        notified until the next notification.
        """
        value = object()
        n = Notifier()
        box = []

        def add_new_waiter(_):
            box.append(n.wait())

        n.wait().addCallback(add_new_waiter)
        n.notify(object())
        self.assertNoResult(box[0])
        n.notify(value)
        self.assertEqual(
            (yield box[0]),
            value,
        )

    @defer.inlineCallbacks
    def test_notify_failure(self):
        """
        If a failure is passed to `Notifier.notify` then the waiters
        are errback'd.
        """
        n = Notifier()
        d = n.wait()
        n.notify(Failure(TestException()))
        with self.assertRaises(TestException):
            yield d

    def test_nonzero_waiters(self):
        """
        If there are waiters, ``Notifier`` evaluates as `True`.
        """
        n = Notifier()
        n.wait()
        self.assertTrue(n)

    def test_nonzero_no_waiters(self):
        """
        If there no waiters, ``Notifier`` evaluates as `False`.
        """
        n = Notifier()
        self.assertFalse(n)

    def test_nonzero_cleared_waiters(self):
        """
        After notifying waiters, ``Notifier`` evaluates as `False`.
        """
        n = Notifier()
        n.wait()
        n.notify(object())
        self.assertFalse(n)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

# This module is left for backward compatibility of old-named worker API.
# It should never be imported by Buildbot.
from __future__ import annotations

from typing import Any

from buildbot_worker.scripts import base
from buildbot_worker.scripts import start
from buildbot_worker.scripts import stop


def restart(config: dict[str, Any]) -> int:
    quiet = config['quiet']
    basedir = config['basedir']

    if not base.isWorkerDir(basedir):
        return 1

    try:
        stop.stopWorker(basedir, quiet)
    except stop.WorkerNotRunning:
        if not quiet:
            print("no old worker process found to stop")
    if not quiet:
        print("now restarting worker process..")

    return start.startWorker(basedir, quiet, config['nodaemon'])



#!/usr/bin/env python
"""
github_buildbot.py is based on git_buildbot.py. Last revised on 2014-02-20.

github_buildbot.py will determine the repository information from the JSON
HTTP POST it receives from github.com and build the appropriate repository.
If your github repository is private, you must add a ssh key to the github
repository for the user who initiated the build on the worker.

This version of github_buildbot.py parses v3 of the github webhook api, with the
"application.vnd.github.v3+json" payload. Configure *only* "push" and/or
"pull_request" events to trigger this webhook.

"""

import hmac
import logging
import os
import re
import sys
from hashlib import sha1
from optparse import OptionParser

from future.utils import iteritems
from twisted.cred import credentials
from twisted.internet import reactor
from twisted.spread import pb
from twisted.web import resource
from twisted.web import server

try:
    import json
except ImportError:
    import simplejson as json


ACCEPTED = 202
BAD_REQUEST = 400
INTERNAL_SERVER_ERROR = 500
OK = 200


class GitHubBuildBot(resource.Resource):
    """
    GitHubBuildBot creates the webserver that responds to the GitHub Service
    Hook.
    """

    isLeaf = True
    master = None
    port = None

    def render_POST(self, request):
        """
        Responds only to POST events and starts the build process

        :arguments:
            request
                the http request object
        """

        # All responses are application/json
        request.setHeader(b"Content-Type", b"application/json")

        content = request.content.read()

        # Verify the message if a secret was provided
        #
        # NOTE: We always respond with '400 BAD REQUEST' if we can't
        # validate the message.  This is done to prevent malicious
        # requests from learning about why they failed to POST data
        # to us.
        if self.secret is not None:
            signature = request.getHeader(b"X-Hub-Signature")

            if signature is None:
                logging.error("Rejecting request.  Signature is missing.")
                request.setResponseCode(BAD_REQUEST)
                return json.dumps({"error": "Bad Request."})

            try:
                hash_type, hexdigest = signature.split(b"=")

            except ValueError:
                logging.error("Rejecting request.  Bad signature format.")
                request.setResponseCode(BAD_REQUEST)
                return json.dumps({"error": "Bad Request."})

            else:
                # sha1 is hard coded into github's source code so it's
                # unlikely this will ever change.
                if hash_type != b"sha1":
                    logging.error("Rejecting request.  Unexpected hash type.")
                    request.setResponseCode(BAD_REQUEST)
                    return json.dumps({"error": "Bad Request."})

                mac = hmac.new(self.secret, msg=content, digestmod=sha1)
                if mac.hexdigest() != hexdigest:
                    logging.error("Rejecting request.  Hash mismatch.")
                    request.setResponseCode(BAD_REQUEST)
                    return json.dumps({"error": "Bad Request."})

        event_type = request.getHeader(b"X-GitHub-Event")
        logging.debug(b"X-GitHub-Event: " + event_type)

        handler = getattr(self, 'handle_' + event_type.decode("ascii"), None)

        if handler is None:
            logging.info("Rejecting request. Received unsupported event %r.", event_type)
            request.setResponseCode(BAD_REQUEST)
            return json.dumps({"error": "Bad Request."})

        try:
            content_type = request.getHeader(b"Content-Type")

            if content_type == b"application/json":
                payload = json.loads(content)
            elif content_type == b"application/x-www-form-urlencoded":
                payload = json.loads(request.args["payload"][0])
            else:
                logging.info(
                    "Rejecting request.  Unknown 'Content-Type', received %r", content_type
                )
                request.setResponseCode(BAD_REQUEST)
                return json.dumps({"error": "Bad Request."})

            logging.debug("Payload: " + payload)
            repo = payload['repository']['full_name']
            repo_url = payload['repository']['html_url']
            changes = handler(payload, repo, repo_url)
            self.send_changes(changes, request)
            return server.NOT_DONE_YET

        except Exception as e:
            logging.exception(e)
            request.setResponseCode(INTERNAL_SERVER_ERROR)
            return json.dumps({"error": str(e)})

    def process_change(self, change, branch, repo, repo_url):
        files = change['added'] + change['removed'] + change['modified']
        who = ""
        if 'username' in change['author']:
            who = change['author']['username']
        else:
            who = change['author']['name']
        if 'email' in change['author']:
            who = "{} <{}>".format(who, change['author']['email'])

        comments = change['message']
        if len(comments) > 1024:
            trim = " ... (trimmed, commit message exceeds 1024 characters)"
            comments = comments[: 1024 - len(trim)] + trim

        info_change = {
            'revision': change['id'],
            'revlink': change['url'],
            'who': who,
            'comments': comments,
            'repository': repo_url,
            'files': files,
            'project': repo,
            'branch': branch,
        }

        if self.category:
            info_change['category'] = self.category

        return info_change

    def handle_ping(self, *_):
        return None

    def handle_push(self, payload, repo, repo_url):
        """
        Consumes the JSON as a python object and actually starts the build.

        :arguments:
            payload
                Python Object that represents the JSON sent by GitHub Service
                Hook.
        """
        changes = None
        refname = payload['ref']

        if self.filter_push_branch:
            if refname != f"refs/heads/{self.filter_push_branch}":
                logging.info(
                    f"Ignoring refname '{refname}': Not a push to branch '{self.filter_push_branch}'"
                )
                return changes

        m = re.match(r"^refs/(heads|tags)/(.+)$", refname)
        if not m:
            logging.info("Ignoring refname '%s': Not a branch or a tag", refname)
            return changes

        refname = m.group(2)

        if payload['deleted'] is True:
            logging.info("%r deleted, ignoring", refname)
        else:
            changes = []
            for change in payload['commits']:
                if (self.head_commit or m.group(1) == 'tags') and change['id'] != payload[
                    'head_commit'
                ]['id']:
                    continue
                changes.append(self.process_change(change, refname, repo, repo_url))
        return changes

    def handle_pull_request(self, payload, repo, repo_url):
        """
        Consumes the JSON as a python object and actually starts the build.

        :arguments:
            payload
                Python Object that represents the JSON sent by GitHub Service
                Hook.
        """
        changes = None

        branch = "refs/pull/{}/head".format(payload['number'])

        if payload['action'] not in ("opened", "synchronize"):
            logging.info("PR %r %r, ignoring", payload['number'], payload['action'])
            return None
        else:
            changes = []

            # Create a synthetic change
            change = {
                'id': payload['pull_request']['head']['sha'],
                'message': payload['pull_request']['body'],
                'timestamp': payload['pull_request']['updated_at'],
                'url': payload['pull_request']['html_url'],
                'author': {
                    'username': payload['pull_request']['user']['login'],
                },
                'added': [],
                'removed': [],
                'modified': [],
            }

            changes.append(self.process_change(change, branch, repo, repo_url))
        return changes

    def send_changes(self, changes, request):
        """
        Submit the changes, if any
        """
        if not changes:
            logging.warning("No changes found")
            request.setResponseCode(OK)
            request.write(json.dumps({"result": "No changes found."}))
            request.finish()
            return

        host, port = self.master.split(':')
        port = int(port)

        if self.auth is not None:
            auth = credentials.UsernamePassword(*self.auth.split(":"))
        else:
            auth = credentials.Anonymous()

        factory = pb.PBClientFactory()
        deferred = factory.login(auth)
        reactor.connectTCP(host, port, factory)
        deferred.addErrback(self.connectFailed, request)
        deferred.addCallback(self.connected, changes, request)

    def connectFailed(self, error, request):
        """
        If connection is failed.  Logs the error.
        """
        logging.error("Could not connect to master: %s", error.getErrorMessage())
        request.setResponseCode(INTERNAL_SERVER_ERROR)
        request.write(json.dumps({"error": "Failed to connect to buildbot master."}))
        request.finish()
        return error

    def addChange(self, _, remote, changei, src='git'):
        """
        Sends changes from the commit to the buildmaster.
        """
        logging.debug("addChange %r, %r", remote, changei)
        try:
            change = changei.next()
        except StopIteration:
            remote.broker.transport.loseConnection()
            return None

        logging.info("New revision: %s", change['revision'][:8])
        for key, value in iteritems(change):
            logging.debug("  %s: %s", key, value)

        change['src'] = src
        deferred = remote.callRemote('addChange', change)
        deferred.addCallback(self.addChange, remote, changei, src)
        return deferred

    def connected(self, remote, changes, request):
        """
        Responds to the connected event.
        """
        # By this point we've connected to buildbot so
        # we don't really need to keep github waiting any
        # longer
        request.setResponseCode(ACCEPTED)
        request.write(json.dumps({"result": "Submitting changes."}))
        request.finish()

        return self.addChange(None, remote, changes.__iter__())


def setup_options():
    """
    The main event loop that starts the server and configures it.
    """
    usage = "usage: %prog [options]"
    parser = OptionParser(usage)

    parser.add_option(
        "-p",
        "--port",
        help="Port the HTTP server listens to for the GitHub Service Hook [default: %default]",
        default=9001,
        type=int,
        dest="port",
    )

    parser.add_option(
        "-m",
        "--buildmaster",
        help="Buildbot Master host and port. ie: localhost:9989 [default: %default]",
        default="localhost:9989",
        dest="buildmaster",
    )

    parser.add_option(
        "--auth",
        help="The username and password, separated by a colon, "
        "to use when connecting to buildbot over the "
        "perspective broker.",
        default="change:changepw",
        dest="auth",
    )

    parser.add_option(
        "--head-commit", action="store_true", help="If set, only trigger builds for commits at head"
    )

    parser.add_option(
        "--secret",
        help="If provided then use the X-Hub-Signature header "
        "to verify that the request is coming from "
        "github. [default: %default]",
        default=None,
        dest="secret",
    )

    parser.add_option(
        "-l",
        "--log",
        help="The absolute path, including filename, to save the "
        "log to [default: %default].  This may also be "
        "'stdout' indicating logs should output directly to "
        "standard output instead.",
        default="github_buildbot.log",
        dest="log",
    )

    parser.add_option(
        "-L",
        "--level",
        help="The logging level: debug, info, warn, error, fatal [default: %default]",
        default='warn',
        dest="level",
        choices=("debug", "info", "warn", "error", "fatal"),
    )

    parser.add_option(
        "-g",
        "--github",
        help="The github server.  Changing this is useful if"
        " you've specified a specific HOST handle in "
        "~/.ssh/config for github [default: %default]",
        default='github.com',
        dest="github",
    )

    parser.add_option(
        "--pidfile",
        help="Write the process identifier (PID) to this "
        "file on start. The file is removed on clean "
        "exit. [default: %default]",
        default=None,
        dest="pidfile",
    )

    parser.add_option(
        "--category", help="Category for the build change", default=None, dest="category"
    )

    parser.add_option(
        "--filter-push-branch",
        help="Only trigger builds for pushes to a given branch name.",
        default=None,
        dest="filter_push_branch",
    )

    (options, _) = parser.parse_args()

    if options.auth is not None and ":" not in options.auth:
        parser.error("--auth did not contain ':'")

    if options.pidfile:
        with open(options.pidfile, 'w') as f:
            f.write(str(os.getpid()))

    filename = options.log
    log_format = "%(asctime)s - %(levelname)s - %(message)s"
    if options.log != "stdout":
        logging.basicConfig(
            filename=filename, format=log_format, level=logging._levelNames[options.level.upper()]
        )
    else:
        logging.basicConfig(
            format=log_format,
            handlers=[logging.StreamHandler(stream=sys.stdout)],
            level=logging._levelNames[options.level.upper()],
        )

    return options


def run_hook(options):
    github_bot = GitHubBuildBot()
    github_bot.github = options.github
    github_bot.master = options.buildmaster
    github_bot.secret = options.secret
    github_bot.auth = options.auth
    github_bot.head_commit = options.head_commit
    github_bot.category = options.category
    github_bot.filter_push_branch = options.filter_push_branch

    site = server.Site(github_bot)
    reactor.listenTCP(options.port, site)

    reactor.run()


def main():
    options = setup_options()

    run_hook(options)


if __name__ == '__main__':
    main()
